import os
import shutil
from pathlib import Path

import pytest
import pytest_asyncio
from hishel import AsyncInMemoryStorage, AsyncRedisStorage

from pynspd import AsyncNspd


@pytest_asyncio.fixture(scope="session")
async def file_cache_api():
    async with AsyncNspd(cache_folder_path=".test.cache") as client:
        yield client
        shutil.rmtree(".test.cache")


@pytest_asyncio.fixture(scope="session")
async def sqlite_cache_api():
    cache_path = (Path.cwd() / ".test.cache.sqlite").resolve()
    async with AsyncNspd(cache_sqlite_url=cache_path) as client:
        yield client
    cache_path.unlink()


@pytest_asyncio.fixture(scope="session")
async def redis_cache_api():
    async with AsyncNspd(
        cache_redis_url=os.environ["REDIS_URL"], trust_env=False
    ) as client:
        yield client
        cache: AsyncRedisStorage = client._cache_storage
        await cache._client.flushdb()


@pytest_asyncio.fixture(scope="session")
async def custom_cache_api():
    async with AsyncNspd(cache_storage=AsyncInMemoryStorage()) as client:
        yield client


async def request(api: AsyncNspd):
    return await api.request(
        "get",
        "/async_api/geoportal/v2/search/geoportal",
        params={
            "query": "77:02:0021001:5304",
            "thematicSearchId": 1,
        },
    )


async def req_assertion(api: AsyncNspd):
    r = await request(api)
    assert not r.extensions["from_cache"]
    r = await request(api)
    assert r.extensions["from_cache"]


@pytest.mark.asyncio(scope="session")
async def test_file_cache_client(file_cache_api: AsyncNspd):
    await req_assertion(file_cache_api)


@pytest.mark.asyncio(scope="session")
async def test_sqlite_cache_client(sqlite_cache_api: AsyncNspd):
    await req_assertion(sqlite_cache_api)


@pytest.mark.asyncio(scope="session")
async def test_redis_cache_client(redis_cache_api: AsyncNspd):
    await req_assertion(redis_cache_api)


@pytest.mark.asyncio(scope="session")
async def test_custom_cache_client(custom_cache_api: AsyncNspd):
    await req_assertion(custom_cache_api)


@pytest.mark.asyncio(scope="session")
async def test_multiple_cache_client():
    with pytest.raises(ValueError):
        async with AsyncNspd(cache_redis_url="foo", cache_sqlite_url="bar") as api:
            await req_assertion(api)



from pynspd import errors
from pynspd._async.api import AsyncNspd
from pynspd._sync.api import Nspd
from pynspd.map_types.enums import ThemeId
from pynspd.schemas import NspdFeature

__all__ = ["AsyncNspd", "Nspd", "NspdFeature", "ThemeId", "errors"]
__version__ = "1.1.7"



import pytest

from pynspd import Nspd, ThemeId
from pynspd.errors import UnknownLayer
from pynspd.schemas import Layer36049Feature, Options36369, Options36383
from pynspd.schemas.base_feature import BaseFeature


def find(cn: str):
    with Nspd() as api:
        feat = api.find(cn)
        assert feat is not None
        return feat


def test_feature_cast():
    """Приведение общего объекта к объекту конкретного слоя"""
    cf = find("77:01:0001044:1033").cast()
    assert isinstance(cf, Layer36049Feature)
    assert len(cf.properties.options.model_dump_human_readable()) > 0


def test_feature_cast_by_category():
    """Приведение объекта из скрытого слоя, но с известной категорией"""
    cf = find("77:01:0001044:2938").cast()
    assert isinstance(cf, BaseFeature)
    assert isinstance(cf.properties.options, Options36383)
    assert len(cf.properties.options.model_dump_human_readable()) > 0


def test_feature_failed_cast_on_hidden_layer():
    """Попытка приведения объекта из неизвестного слоя с неизвестной категорией"""
    with pytest.raises(UnknownLayer):
        find("77:02:0021001:5304").cast()


def test_props_cast():
    """Приведение общих свойств к свойствам конкретного слоя"""
    props = find("77:01:0001044:1033").properties.cast().options
    assert isinstance(props, Options36369)
    assert len(props.model_dump_human_readable()) > 0


def test_props_failed_cast_on_hidden_layer():
    """Попытка приведения свойств из неизвестного слоя"""
    with pytest.raises(UnknownLayer):
        find("77:02:0021001:5304").properties.cast()


def test_feature_repr(api: Nspd):
    f = api.find("77:02:0021001:5304")
    assert repr(f) == "NspdFeature<Машино-места: 77:02:0021001:5304>"


def test_uncasted_non_cn_feature_repr(api: Nspd):
    f = api.find("Останкинская", theme_id=ThemeId.ZONES_AND_TERRITORIES)
    assert repr(f) == "NspdFeature<Зоны с особыми условиями использования территории>"



import dotenv
import pytest
import pytest_asyncio

from pynspd import AsyncNspd, Nspd

dotenv.load_dotenv()


@pytest.fixture(scope="session")
def api():
    with Nspd() as client:
        yield client


@pytest_asyncio.fixture(scope="session")
async def async_api():
    async with AsyncNspd() as client:
        yield client



# autogenerated
from datetime import datetime
from typing import Optional, Union

from pynspd.schemas.base_feature import BaseFeature
from pynspd.schemas.geometries import (
    LineString,
    MultiLineString,
    MultiPoint,
    MultiPolygon,
    Point,
    Polygon,
)
from pynspd.schemas.layer_configs import LayerNode
from pynspd.schemas.properties import NspdProperties, OptionProperties


class Options36992(OptionProperties):
    title_key = "brd_nmb"
    type_boundary_value: Optional[str] = None
    """Вид"""
    registration_date: Optional[datetime] = None
    """Дата присвоения"""
    brd_nmb: Optional[str] = None
    """Реестровый номер границы"""
    cadastral_district: Optional[str] = None
    """Кадастровый район"""
    name: Optional[str] = None
    """Наименование"""
    doc: Optional[str] = None
    """Реквизиты решения об установлении (изменении) границ"""
    document_name: Optional[str] = None
    """Наименование документа"""
    document_number: Optional[str] = None
    """Номер"""
    document_date: Optional[datetime] = None
    """Дата"""
    document_issuer: Optional[str] = None
    """Орган власти, организация выдавшие документ"""


class Options38030(OptionProperties):
    title_key = "name"
    obj_kind_value: Optional[str] = None
    """Тип"""
    name: Optional[str] = None
    """Наименование"""
    cad_num: Optional[str] = None
    """Учетный номер"""
    cnt_kr: Optional[float] = None
    """Количество кадастровых районов"""
    cnt_kk: Optional[float] = None
    """Количество кадастровых кварталов"""
    cnt_land: Optional[float] = None
    """Количество земельных участков"""
    cnt_land_geom: Optional[float] = None
    """c границами"""
    cnt_land_not_geom: Optional[float] = None
    """без границ"""
    cnt_oks: Optional[float] = None
    """Количество объектов капитального строительства"""
    cnt_oks_geom: Optional[float] = None
    """с границами"""
    cnt_oks_not_geom: Optional[float] = None
    """без границ"""
    cnt_enk: Optional[float] = None
    """Количество единых недвижимых комплексов"""
    cnt_pik: Optional[float] = None
    """Количество предприятий как имущественных комплексов"""


class Options36368(OptionProperties):
    title_key = "cad_num"
    land_record_type: Optional[str] = None
    """Вид объекта недвижимости"""
    land_record_subtype: Optional[str] = None
    """Вид земельного участка"""
    land_record_reg_date: Optional[datetime] = None
    """Дата присвоения"""
    cad_num: Optional[str] = None
    """Кадастровый номер"""
    quarter_cad_number: Optional[str] = None
    """Кадастровый квартал"""
    readable_address: Optional[str] = None
    """Адрес"""
    specified_area: Optional[float] = None
    """Площадь уточненная (кв. м)"""
    declared_area: Optional[float] = None
    """Площадь декларированная (кв. м)"""
    area: Optional[float] = None
    """Площадь (кв. м)"""
    status: Optional[str] = None
    """Статус"""
    land_record_category_type: Optional[str] = None
    """Категория земель"""
    permitted_use_established_by_document: Optional[str] = None
    """Вид разрешенного использования"""
    ownership_type: Optional[str] = None
    """Форма собственности"""
    cost_value: Optional[float] = None
    """Кадастровая стоимость (руб.)"""
    cost_index: Optional[float] = None
    """Удельный показатель кадастровой стоимости (руб./кв. м)"""


class Options36369(OptionProperties):
    title_key = "cad_num"
    build_record_type_value: Optional[str] = None
    """Вид объекта недвижимости"""
    build_record_registration_date: Optional[datetime] = None
    """Дата присвоения"""
    cad_num: Optional[str] = None
    """Кадастровый номер"""
    quarter_cad_number: Optional[str] = None
    """Кадастровый квартал"""
    readable_address: Optional[str] = None
    """Адрес"""
    building_name: Optional[str] = None
    """Наименование"""
    purpose: Optional[str] = None
    """Назначение"""
    build_record_area: Optional[float] = None
    """Площадь общая (кв. м)"""
    status: Optional[str] = None
    """Статус"""
    ownership_type: Optional[str] = None
    """Форма собственности"""
    cost_value: Optional[float] = None
    """Кадастровая стоимость (руб.)"""
    cost_index: Optional[float] = None
    """Удельный показатель кадастровой стоимости (руб./ кв. м)"""
    floors: Optional[str] = None
    """Количество этажей (в том числе подземных)"""
    underground_floors: Optional[str] = None
    """Количество подземных этажей"""
    materials: Optional[str] = None
    """Материал стен"""
    year_built: Optional[str] = None
    """Завершение строительства (г.)"""
    year_commisioning: Optional[str] = None
    """Ввод в эксплуатацию (г.)"""
    cultural_heritage_object: Optional[str] = None
    """Сведения о включении объекта недвижимости в единый государственный реестр объектов культурного наследия (памятников истории и культуры) народов Российской Федерации"""
    united_cad_number: Optional[str] = None
    """Кадастровый номер единого недвижимого комплекса, в состав которого входит здание"""
    facility_cad_number: Optional[str] = None
    """Кадастровый номер предприятия как имущественного комплекса, в состав которого входит здание (имущественное право на здание)"""


class Options37148(OptionProperties):
    title_key = "brd_nmb"
    type_boundary_value: Optional[str] = None
    """Вид"""
    registration_date: Optional[datetime] = None
    """Дата присвоения"""
    brd_nmb: Optional[str] = None
    """Реестровый номер границы"""
    cadastral_district: Optional[str] = None
    """Кадастровый район"""
    name: Optional[str] = None
    """Наименование"""
    doc: Optional[str] = None
    """Реквизиты решения об установлении (изменении) границ"""
    document_name: Optional[str] = None
    """Наименование документа"""
    document_number: Optional[str] = None
    """Номер"""
    document_date: Optional[datetime] = None
    """Дата"""
    document_issuer: Optional[str] = None
    """Орган власти, организация выдавшие документ"""


class Options36381(OptionProperties):
    title_key = "cad_num"
    obj_kind_value: Optional[str] = None
    """Тип"""
    name: Optional[str] = None
    """Наименование"""
    cad_num: Optional[str] = None
    """Учетный номер"""
    cad_num_ko: Optional[str] = None
    """Кадастровый округ"""
    cad_num_kr: Optional[str] = None
    """Кадастровый район"""
    cnt_land: Optional[float] = None
    """Количество земельных участков"""
    cnt_land_geom: Optional[float] = None
    """с границами"""
    cnt_land_not_geom: Optional[float] = None
    """без границ"""
    cnt_oks: Optional[float] = None
    """Количество объектов капитального строительства"""
    cnt_oks_geom: Optional[float] = None
    """с границами"""
    cnt_oks_not_geom: Optional[float] = None
    """без границ"""
    cnt_enk: Optional[float] = None
    """Количество единых недвижимых комплексов"""
    cnt_pik: Optional[float] = None
    """Количество предприятий как имущественных комплексов"""


class Options36382(OptionProperties):
    title_key = "cad_num"
    obj_kind_value: Optional[str] = None
    """Тип"""
    name: Optional[str] = None
    """Наименование"""
    cad_num: Optional[str] = None
    """Учетный номер"""
    cad_num_ko: Optional[str] = None
    """Кадастровый округ"""
    cnt_kk: Optional[float] = None
    """Количество кадастровых кварталов"""
    cnt_land: Optional[float] = None
    """Количество земельных участков"""
    cnt_land_geom: Optional[float] = None
    """с границами"""
    cnt_land_not_geom: Optional[float] = None
    """без границ"""
    cnt_oks: Optional[float] = None
    """Количество объектов капитального строительства"""
    cnt_oks_geom: Optional[float] = None
    """с границами"""
    cnt_oks_not_geom: Optional[float] = None
    """без границ"""
    cnt_enk: Optional[float] = None
    """Количество единых недвижимых комплексов"""
    cnt_pik: Optional[float] = None
    """Количество предприятий как имущественных комплексов"""


class Options38943(OptionProperties):
    title_key = "address"
    id: Optional[str] = None
    """ID"""
    doc_guid: Optional[str] = None
    """GUID"""
    doc_type: Optional[str] = None
    """Наименование вида документа об утверждении схемы расположения земельного участка"""
    doc_gov: Optional[str] = None
    """наименование уполномоченного органа"""
    doc_date: Optional[datetime] = None
    """дата"""
    doc_number: Optional[str] = None
    """номер"""
    cond_numbe: Optional[str] = None
    """Условный номер образуемого земельного участка"""
    shape_area: Optional[float] = None
    """Площадь (кв. м)"""
    address: Optional[str] = None
    """Адрес или иное описание местоположения земельного участка"""
    kvartal_id: Optional[str] = None
    """Кадастровый номер земельного участка, из которого в соответствии со схемой расположения земельного участка предусмотрено образование земельного участка"""
    name_zone: Optional[str] = None
    """Территориальная зона, в границах которой образуется земельный участок"""
    category_code: Optional[str] = None
    """Категория земель образуемого земельного"""
    utilization: Optional[str] = None
    """Вид разрешенного использования образуемого земельного участка"""


class Options36383(OptionProperties):
    title_key = "cad_number"
    object_type_value: Optional[str] = None
    """Вид объекта недвижимости"""
    registration_date: Optional[datetime] = None
    """Дата присвоения"""
    cad_number: Optional[str] = None
    """Кадастровый номер"""
    quarter_cad_number: Optional[str] = None
    """Кадастровый квартал"""
    address_readable_address: Optional[str] = None
    """Адрес"""
    params_name: Optional[str] = None
    """Наименование"""
    params_purpose: Optional[str] = None
    """Назначение"""
    params_height: Optional[float] = None
    """высота (м.)"""
    params_depth: Optional[float] = None
    """глубина (м.)"""
    params_occurence_depth: Optional[float] = None
    """глубина залегания (м.)"""
    params_extension: Optional[float] = None
    """протяженность (м.)"""
    params_volume: Optional[float] = None
    """объем (куб. м)"""
    params_built_up_area: Optional[float] = None
    """Площадь застройки (кв. м)"""
    params_area: Optional[float] = None
    """Площадь (кв. м)"""
    object_previously_posted: Optional[str] = None
    """Статус"""
    ownership_type: Optional[str] = None
    """Форма собственности"""
    cost_value: Optional[float] = None
    """Кадастровая стоимость (руб.)"""
    cost_index: Optional[float] = None
    """Удельный показатель кадастровой стоимости (руб./кв. м)"""
    params_floors: Optional[str] = None
    """Количество этажей (в том числе подземных)"""
    params_underground_floors: Optional[str] = None
    """Количество подземных этажей"""
    params_year_built: Optional[str] = None
    """Завершение строительства"""
    params_year_commisioning: Optional[str] = None
    """Ввод в эксплуатацию"""
    cultural_heritage_object: Optional[str] = None
    """Сведения о включении объекта недвижимости в единый государственный реестр объектов культурного наследия (памятников истории и культуры) народов Российской Федерации"""
    united_cad_number: Optional[str] = None
    """Кадастровый номер единого недвижимого комплекса, в состав которого входит сооружение"""
    facility_cad_number: Optional[str] = None
    """Кадастровый номер предприятия как имущественного комплекса, в состав которого сооружение (имущественное право на здание) входит"""


class Options36384(OptionProperties):
    title_key = "cad_num"
    object_under_construction_record_record_type_value: Optional[str] = None
    """Вид объекта недвижимости"""
    registration_date: Optional[datetime] = None
    """Дата присвоения"""
    cad_num: Optional[str] = None
    """Кадастровый номер"""
    quarter_cad_number: Optional[str] = None
    """Кадастровый квартал"""
    readable_address: Optional[str] = None
    """Адрес"""
    object_under_construction_record_name: Optional[str] = None
    """Наименование"""
    height: Optional[float] = None
    """Высота (м.)"""
    depth: Optional[float] = None
    """Глубина (м.)"""
    occurence_depth: Optional[float] = None
    """Глубина залегания (м.)"""
    extension: Optional[float] = None
    """Протяжённость (м.)"""
    volume: Optional[float] = None
    """Объем (куб. м)"""
    built_up_area: Optional[float] = None
    """Площадь застройки (кв. м)"""
    common_data_status: Optional[str] = None
    """Статус"""
    ownership_type: Optional[str] = None
    """Форма собственности"""
    cost_value: Optional[float] = None
    """Кадастровая стоимость (руб.)"""
    cost_index: Optional[float] = None
    """Удельный показатель кадастровой стоимости (руб./ кв. м)"""
    degree_readiness: Optional[str] = None
    """Степень готовности ОНС в процентах (%)"""
    purpose: Optional[str] = None
    """Проектируемое значение основной характеристики"""
    facility_cad_number: Optional[str] = None
    """Кадастровый номер предприятия как имущественного комплекса, если объект незавершенного строительства (имущественное право на объект незавершенного строительства) входит в состав предприятия как имущественного комплекса"""


class Options38942(OptionProperties):
    title_key = "brd_nmb"
    content_restrict_encumbrances: Optional[str] = None
    """Ограничения"""
    doc: Optional[str] = None
    """Реквизиты решения об утверждении документации по планировке территории"""
    DOC_TYPE: Optional[str] = None
    """документ"""
    DOC_NUMBER: Optional[str] = None
    """номер"""
    DOC_DATE: Optional[datetime] = None
    """дата"""
    DOC_GOV: Optional[str] = None
    """орган власти, организация выдавшие документ"""


class Options445347(OptionProperties):
    title_key = "type"
    cn: Optional[str] = None
    """Кадастровый или условный номер"""
    kvartal_cn: Optional[str] = None
    """Кадастровый квартал"""
    reg_name: Optional[str] = None
    """Субъект Российской Федерации"""
    mun: Optional[str] = None
    """Муниципальное образование"""
    np: Optional[str] = None
    """Населенный пункт"""
    address: Optional[str] = None
    """Адрес"""
    area: Optional[float] = None
    """Площадь"""
    category_text: Optional[str] = None
    """Категория земель"""
    util_by_doc: Optional[str] = None
    """Вид разрешенного использования"""
    right: Optional[str] = None
    """Форма собственности"""
    potential: Optional[str] = None
    """Потенциал использования"""
    has_oks: Optional[str] = None
    """Объекты капитального строительства"""
    has_network: Optional[str] = None
    """Инженерные сети"""
    develop_build: Optional[str] = None
    """Входит в территорию, в отношении которой возможно (планируется) комплексное развитие территории жилой застройки"""
    privileges_peo: Optional[str] = None
    """Предназначена для предоставления льготным категориям граждан"""
    privileges: Optional[str] = None
    """Льготная категория"""
    dom_rf: Optional[str] = None
    """Агентом Российской Федерации в отношении земельного участка выступает АО «ДОМ. РФ»"""
    uo_name: Optional[str] = None
    """Наименование уполномоченного органа"""
    uo_address: Optional[str] = None
    """Адрес уполномоченного органа"""
    uo_phone: Optional[str] = None
    """Телефон уполномоченного органа"""
    uo_url: Optional[str] = None
    """Сайт в сети Интернет"""
    uo_email: Optional[str] = None
    """e-mail"""
    inves_portal: Optional[str] = None
    """Инвестиционный портал региона"""


class Options38944(OptionProperties):
    schema1: Optional[str] = None
    """Тип"""
    name: Optional[str] = None
    """Наименование объекта"""
    part_type: Optional[str] = None
    """Тип части"""
    building: Optional[str] = None
    """Наличие объектов капитального строительства"""


class Options37158(OptionProperties):
    title_key = "nominal_number"
    nominal_number: Optional[str] = None
    """Условный номер образуемого земельного участка"""
    area: Optional[str] = None
    """Площадь образуемого земельного участка (кв. м)"""
    land_use_text: Optional[str] = None
    """Вид разрешенного использования образуемого земельного участка"""
    brd_nmb: Optional[str] = None
    """номер"""
    registration_date: Optional[datetime] = None
    """дата"""
    document_name: Optional[str] = None
    """документ"""
    document_number: Optional[str] = None
    """номер"""
    document_date: Optional[datetime] = None
    """дата"""
    document_issuer: Optional[str] = None
    """орган власти, организация выдавшие документ"""


class Options278838(OptionProperties):
    type_boundary_value: Optional[str] = None
    """Вид"""
    registration_date: Optional[datetime] = None
    """Дата присвоения"""
    reg_numb_border: Optional[str] = None
    """Реестровый номер границы"""
    old_account_number: Optional[str] = None
    """Учетный номер"""
    cadastral_district: Optional[str] = None
    """Кадастровый район"""
    name_by_doc: Optional[str] = None
    """Наименование зоны"""
    document_name: Optional[str] = None
    """документ"""
    document_number: Optional[str] = None
    """номер"""
    document_date: Optional[str] = None
    """дата"""
    document_issuer: Optional[str] = None
    """орган власти, организация выдавшие документ"""


class Options38967(OptionProperties):
    title_key = "process_ty_title"
    sub_rf: Optional[str] = None
    """Субъект РФ"""
    cad_distr: Optional[str] = None
    """Кадастровый округ"""
    district: Optional[str] = None
    """Район"""
    process_ty_title: Optional[str] = None
    """Вид и степень развития негативного процесса"""
    index_map: Optional[str] = None
    """Индекс"""
    shape_area: Optional[float] = None
    """Площадь проявления негативного процесса, га"""
    zone_proya: Optional[str] = None
    """Зона проявления"""
    process_ty: Optional[str] = None
    """Тип"""
    recommend: Optional[str] = None
    """Рекомендации по использованию земель"""
    year: Optional[str] = None
    """Год обследования"""


class Options39228(OptionProperties):
    title_key = "terr_name"
    kkaccountnumber: Optional[str] = None
    """Учетный номер кадастрового квартала"""
    stateKKR: Optional[str] = None
    """Статус выполнения комплексных кадастровых работ"""
    contract: Optional[datetime] = None
    """Дата контракта на выполнение комплексных кадастровых работ"""
    contractNum: Optional[str] = None
    """Номер контракта на выполнение комплексных кадастровых работ"""
    krstartdate: Optional[datetime] = None
    """Дата начала комплексных кадастровых работ"""
    krenddate: Optional[datetime] = None
    """Дата окончания комплексных кадастровых работ"""
    krcustomername: Optional[str] = None
    """Наименование заказчика комплексных кадастровых работ"""
    krcustomeraddress: Optional[str] = None
    """Адрес заказчика комплексных кадастровых работ"""
    krcustomeremail: Optional[str] = None
    """Адрес электронной почты заказчика комплексных кадастровых работ"""
    krcustomerphone: Optional[str] = None
    """Телефон заказчика комплексных кадастровых работ"""
    krcontractorname: Optional[str] = None
    """Наименование юридического лица, с которым заключён контракт на выполнение комплексных кадастровых работ"""
    krsurveyorname: Optional[str] = None
    """ФИО кадастрового инженера"""
    krsurveyoraddress: Optional[str] = None
    """Адрес кадастрового инженера"""
    krsurveyoremail: Optional[str] = None
    """Адрес электронной почты кадастрового инженера"""
    krsurveyorphone: Optional[str] = None
    """Телефон кадастрового инженера"""
    krsurveyoratt: Optional[str] = None
    """Номер аттестата кадастрового инженера"""
    krsurveyorattdate: Optional[datetime] = None
    """Дата выдачи аттестата кадастрового инженера"""
    krsurveyorsroname: Optional[str] = None
    """Наименование саморегулируемой организации, членом которой является кадастровый инженер"""
    typeKKR: Optional[str] = None
    """Время выполнения работ/Место проведения работ/Виды работ"""
    krsurveyoraddress: Optional[str] = None
    """Адрес для предоставления документов о ранее учтенных земельных участках"""


class Options38979(OptionProperties): ...


class Options38981(OptionProperties):
    title_key = "Кадастровый номер"
    auc_thrf: Optional[str] = None
    """Ссылка на аукцион"""


class Options36940(OptionProperties):
    title_key = "name_by_doc"
    type_boundary_value: Optional[str] = None
    """Вид"""
    type_zone: Optional[str] = None
    """Тип зоны"""
    registration_date: Optional[datetime] = None
    """Дата присвоения"""
    reg_numb_border: Optional[str] = None
    """Реестровый номер границы"""
    cadastral_district: Optional[str] = None
    """Кадастровый район"""
    old_account_number: Optional[str] = None
    """Учетный номер"""
    name_by_doc: Optional[str] = None
    """Наименование"""
    content_restrict_encumbrances: Optional[str] = None
    """Ограничение"""
    doc: Optional[str] = None
    """Реквизиты решения об установлении (изменении) ЗОУИТ"""
    legal_act_document_name: Optional[str] = None
    """документ"""
    legal_act_document_number: Optional[str] = None
    """номер"""
    legal_act_document_date: Optional[datetime] = None
    """дата"""
    legal_act_document_issuer: Optional[str] = None
    """орган власти, организация выдавшие документ"""


class Options36941(OptionProperties):
    title_key = "name_by_doc"
    type_boundary_value: Optional[str] = None
    """Вид"""
    registration_date: Optional[datetime] = None
    """Дата присвоения"""
    reg_numb_border: Optional[str] = None
    """Реестровый номер границы"""
    cadastral_district: Optional[str] = None
    """Кадастровый район"""
    old_account_number: Optional[str] = None
    """Учетный номер"""
    name_by_doc: Optional[str] = None
    """Наименование"""
    description: Optional[str] = None
    """Описание"""
    document_name: Optional[str] = None
    """Реквизиты правовых актов об установлении или изменении границ"""
    legal_act_document_name: Optional[str] = None
    """Наименование документа"""
    legal_act_document_number: Optional[str] = None
    """Номер"""
    legal_act_document_date: Optional[datetime] = None
    """Дата"""
    legal_act_document_issuer: Optional[str] = None
    """Орган власти, организация выдавшие документ"""


class Options36942(OptionProperties):
    title_key = "name_by_doc"
    type_boundary_value: Optional[str] = None
    """Вид"""
    registration_date: Optional[datetime] = None
    """Дата присвоения"""
    reg_numb_border: Optional[str] = None
    """Реестровый номер границы"""
    cadastral_district: Optional[str] = None
    """Кадастровый район"""
    old_account_number: Optional[str] = None
    """Учетный номер"""
    name_by_doc: Optional[str] = None
    """Наименование"""
    document_name: Optional[str] = None
    """Реквизиты решения об установлении (изменении) границ"""
    legal_act_document_name: Optional[str] = None
    """Наименование документа"""
    legal_act_document_number: Optional[str] = None
    """Номер"""
    legal_act_document_date: Optional[datetime] = None
    """Дата"""
    legal_act_document_issuer: Optional[str] = None
    """Орган власти, организация выдавшие документ"""


class Options36943(OptionProperties):
    title_key = "name_by_doc"
    type_boundary_value: Optional[str] = None
    """Вид"""
    registration_date: Optional[datetime] = None
    """Дата присвоения"""
    reg_numb_border: Optional[str] = None
    """Реестровый номер границы"""
    cadastral_district: Optional[str] = None
    """Кадастровый район"""
    name_by_doc: Optional[str] = None
    """Наименование"""
    document_name: Optional[str] = None
    """Реквизиты решения об установлении или изменении границ"""
    legal_act_document_name: Optional[str] = None
    """Наименование документа"""
    legal_act_document_number: Optional[str] = None
    """Номер"""
    legal_act_document_date: Optional[datetime] = None
    """Дата"""
    legal_act_document_issuer: Optional[str] = None
    """Орган власти, организация выдавшие документ"""


class Options36945(OptionProperties):
    title_key = "name_by_doc"
    type_boundary_value: Optional[str] = None
    """Вид"""
    registration_date: Optional[datetime] = None
    """Дата присвоения"""
    reg_numb_border: Optional[str] = None
    """Реестровый номер границы"""
    cadastral_district: Optional[str] = None
    """Кадастровый район"""
    old_account_number: Optional[str] = None
    """Учетный номер"""
    name_by_doc: Optional[str] = None
    """Наименование"""
    document_name: Optional[str] = None
    """Реквизиты решения об установлении (изменении) границ"""
    legal_act_document_name: Optional[str] = None
    """Наименование документа"""
    legal_act_document_number: Optional[str] = None
    """Номер"""
    legal_act_document_date: Optional[datetime] = None
    """Дата"""
    legal_act_document_issuer: Optional[str] = None
    """Орган власти, организация выдавшие документ"""


class Options36946(OptionProperties):
    title_key = "name_by_doc"
    type_boundary_value: Optional[str] = None
    """Вид"""
    type_zone: Optional[str] = None
    """Тип зоны"""
    registration_date: Optional[datetime] = None
    """Дата присвоения"""
    reg_numb_border: Optional[str] = None
    """Реестровый номер границы"""
    cadastral_district: Optional[str] = None
    """Кадастровый район"""
    old_account_number: Optional[str] = None
    """Учетный номер"""
    name_by_doc: Optional[str] = None
    """Наименование"""
    content_restrict_encumbrances: Optional[str] = None
    """Ограничение"""
    doc: Optional[str] = None
    """Реквизиты решения об установлении (изменении) территориальной зоны"""
    legal_act_document_name: Optional[str] = None
    """документ"""
    legal_act_document_number: Optional[str] = None
    """номер"""
    legal_act_document_date: Optional[datetime] = None
    """дата"""
    legal_act_document_issuer: Optional[str] = None
    """орган власти, организация выдавшие документ"""
    permitted_uses_name: Optional[str] = None
    """Вид разрешенного использования"""


class Options36947(OptionProperties):
    title_key = "name_by_doc"
    type_boundary_value: Optional[str] = None
    """Вид"""
    registration_date: Optional[datetime] = None
    """Дата присвоения"""
    reg_numb_border: Optional[str] = None
    """Реестровый номер границы"""
    cadastral_district: Optional[str] = None
    """Кадастровый район"""
    name_by_doc: Optional[str] = None
    """Наименование"""
    content_restrict_encumbrances: Optional[str] = None
    """Ограничение"""
    document_name: Optional[str] = None
    """Реквизиты решения об установлении или изменении границ"""
    legal_act_document_name: Optional[str] = None
    """Наименование документа"""
    legal_act_document_number: Optional[str] = None
    """Номер"""
    legal_act_document_date: Optional[datetime] = None
    """Дата"""
    legal_act_document_issuer: Optional[str] = None
    """Орган власти, организация выдавшие документ"""


class Options38996(OptionProperties):
    title_key = "name"
    type: Optional[str] = None
    """Вид"""
    brd_nmb: Optional[str] = None
    """Реестровый номер"""
    registration_date: Optional[datetime] = None
    """Дата постановки на учет / регистрации"""
    reg_code: Optional[str] = None
    """Идентификационный номер"""
    name: Optional[str] = None
    """Наименование"""
    doc: Optional[str] = None
    """Реквизиты правовых актов о согласовании и об утверждении изменения границ"""
    document_name: Optional[str] = None
    """документ"""
    document_number: Optional[str] = None
    """номер"""
    document_date: Optional[datetime] = None
    """дата"""
    document_issuer: Optional[str] = None
    """орган власти, организация выдавшие документ"""


class Options445397(OptionProperties):
    title_key = "type"
    cn: Optional[str] = None
    """Кадастровый или условный номер"""
    kvartal_cn: Optional[str] = None
    """Кадастровый квартал"""
    reg_name: Optional[str] = None
    """Субъект Российской Федерации"""
    mun: Optional[str] = None
    """Муниципальное образование"""
    np: Optional[str] = None
    """Населенный пункт"""
    address: Optional[str] = None
    """Адрес"""
    area: Optional[float] = None
    """Площадь, м"""
    category_text: Optional[str] = None
    """Категория земель"""
    terzone_name_by_pzz: Optional[str] = None
    """Территориальная зона"""
    util_by_doc: Optional[str] = None
    """Вид разрешенного использования"""
    right: Optional[str] = None
    """Форма собственности"""
    cad_cost: Optional[str] = None
    """Кадастровая стоимость"""
    power_supply: Optional[str] = None
    """Электроснабжение"""
    water_supply: Optional[str] = None
    """Водоснабжение"""
    gas_supply: Optional[str] = None
    """Газоснабжение"""
    sewerage: Optional[str] = None
    """Канализация"""
    heat_supply: Optional[str] = None
    """Теплоснабжение"""
    poi: Optional[str] = None
    """Наименование ближайшего объекта туристского интереса"""
    poi_cn: Optional[str] = None
    """Кадастровый или реестровый номер объекта туристского интереса"""
    distance_to_near_tour_object: Optional[float] = None
    """Расстояние до ближайшего объекта туристского интереса (км)"""
    id_oti: Optional[str] = None
    """Идентификационный номер объекта туристского интереса"""
    distance_to_near_fed_road: Optional[float] = None
    """Расстояние до ближайшей федеральной трассы (км)"""
    distance_to_near_hard_road: Optional[float] = None
    """Расстояние до ближайшей дороги с твердым покрытием (км)"""
    distance_to_near_district: Optional[float] = None
    """Расстояние до ближайшего центра муниципального района или городского округа (км)"""
    distance_to_near_locality: Optional[float] = None
    """Расстояние до ближайшего населенного пункта (км)"""
    distance_to_near_public_catering: Optional[float] = None
    """Расстояние до ближайшего объекта общественного питания (км)"""
    distance_to_near_health_care: Optional[float] = None
    """Расстояние до ближайшего объекта здравоохранения (км)"""
    distance_to_near_public_transport: Optional[float] = None
    """Расстояние до ближайшей остановки общественного транспорта (км)"""
    distance_to_near_railway_station: Optional[float] = None
    """Расстояние до ближайшей железнодорожной станции (км)"""
    distance_to_near_airport: Optional[float] = None
    """Расстояние до ближайшего аэропорта (км)"""
    additional_info: Optional[str] = None
    """Инвестиционный портал региона"""
    tour_object_description: Optional[str] = None
    """Информационной ресурс, содержащий информацию об объекте туристского интереса"""
    uo_name: Optional[str] = None
    """Наименование уполномоченного органа"""
    uo_phone: Optional[str] = None
    """Телефон уполномоченного органа"""
    uo_address: Optional[str] = None
    """Адрес уполномоченного органа"""
    uo_email: Optional[str] = None
    """e-mail"""


class Options38998(OptionProperties):
    title_key = "brd_nmb"
    type: Optional[str] = None
    """Вид"""
    brd_nmb: Optional[str] = None
    """Реестровый номер"""


class Options38999(OptionProperties):
    title_key = "brd_nmb"
    type: Optional[str] = None
    """Вид"""
    brd_nmb: Optional[str] = None
    """Реестровый номер"""
    date_reg: Optional[datetime] = None
    """Дата постановки на учет / регистрации"""
    name: Optional[str] = None
    """Наименование"""
    doc: Optional[str] = None
    """Реквизиты правовых актов о согласовании и об утверждении изменения границ"""
    legal_act_document_name: Optional[str] = None
    """документ"""
    legal_act_document_number: Optional[str] = None
    """номер"""
    legal_act_document_date: Optional[str] = None
    """дата"""
    legal_act_document_issuer: Optional[str] = None
    """орган власти, организация выдавшие документ"""


class Options39000(OptionProperties):
    title_key = "brd_nmb"
    type: Optional[str] = None
    """Вид"""
    brd_nmb: Optional[str] = None
    """Реестровый номер"""
    date_reg: Optional[datetime] = None
    """Дата постановки на учет / регистрации"""
    name: Optional[str] = None
    """Наименование"""
    doc: Optional[str] = None
    """Реквизиты правовых актов о согласовании и об утверждении изменения границ"""
    legal_act_document_name: Optional[str] = None
    """документ"""
    legal_act_document_number: Optional[str] = None
    """номер"""
    legal_act_document_date: Optional[str] = None
    """дата"""
    legal_act_document_issuer: Optional[str] = None
    """орган власти, организация выдавшие документ"""


class Options36948(OptionProperties):
    title_key = "name_by_doc"
    type_boundary_value: Optional[str] = None
    """Вид"""
    registration_date: Optional[datetime] = None
    """Дата присвоения"""
    reg_numb_border: Optional[str] = None
    """Реестровый номер границы"""
    cadastral_district: Optional[str] = None
    """Кадастровый район"""
    old_account_number: Optional[str] = None
    """Учетный номер"""
    name_by_doc: Optional[str] = None
    """Наименование"""
    content_restrict_encumbrances: Optional[str] = None
    """Ограничение"""
    document_name: Optional[str] = None
    """Реквизиты решения об установлении (изменении) границ"""
    legal_act_document_name: Optional[str] = None
    """Наименование документа"""
    legal_act_document_number: Optional[str] = None
    """Номер"""
    legal_act_document_date: Optional[str] = None
    """Дата"""
    legal_act_document_issuer: Optional[str] = None
    """Орган власти, организация выдавшие документ"""


class Options36829(OptionProperties):
    title_key = "name"
    type: Optional[str] = None
    """Вид"""
    brd_nmb: Optional[str] = None
    """Реестровый номер"""
    registration_date: Optional[datetime] = None
    """Дата постановки на учет / регистрации"""
    reg_code: Optional[str] = None
    """Идентификационный номер"""
    name: Optional[str] = None
    """Наименование"""
    doc: Optional[str] = None
    """Реквизиты правовых актов о согласовании и об утверждении изменения границ"""
    document_name: Optional[str] = None
    """документ"""
    document_number: Optional[str] = None
    """номер"""
    document_date: Optional[str] = None
    """дата"""
    document_issuer: Optional[str] = None
    """орган власти, организация выдавшие документ"""


class Options36830(OptionProperties):
    title_key = "name"
    type: Optional[str] = None
    """Вид"""
    brd_nmb: Optional[str] = None
    """Реестровый номер"""
    registration_date: Optional[datetime] = None
    """Дата постановки на учет / регистрации"""
    reg_code: Optional[str] = None
    """Идентификационный номер"""
    name: Optional[str] = None
    """Наименование"""
    doc: Optional[str] = None
    """Реквизиты правовых актов о согласовании и об утверждении изменения границ"""
    document_name: Optional[str] = None
    """документ"""
    document_number: Optional[str] = None
    """номер"""
    document_date: Optional[str] = None
    """дата"""
    document_issuer: Optional[str] = None
    """орган власти, организация выдавшие документ"""


class Options36832(OptionProperties):
    title_key = "name"
    type: Optional[str] = None
    """Вид"""
    brd_nmb: Optional[str] = None
    """Реестровый номер"""
    registration_date: Optional[datetime] = None
    """Дата постановки на учет / регистрации"""
    reg_code: Optional[str] = None
    """Идентификационный номер"""
    name: Optional[str] = None
    """Наименование"""
    doc: Optional[str] = None
    """Реквизиты правовых актов о согласовании и об утверждении изменения границ"""
    document_name: Optional[str] = None
    """документ"""
    document_number: Optional[str] = None
    """номер"""
    document_date: Optional[datetime] = None
    """дата"""
    document_issuer: Optional[str] = None
    """орган власти, организация выдавшие документ"""


class Options445547(OptionProperties):
    title_key = "id"
    name: Optional[str] = None
    """Наименование объекта туристского интереса"""
    cn: Optional[str] = None
    """Кадастровый номер (при наличии)"""
    address: Optional[str] = None
    """Адрес объекта туристского интереса"""
    type: Optional[str] = None
    """Тип объекта туристского интереса"""
    vid_oopt: Optional[str] = None
    """Вид особо охраняемой природной территории"""
    unesko: Optional[str] = None
    """Наличие у объекта туристского интереса статуса ЮНЕСКО"""
    vid_okn: Optional[str] = None
    """Вид объекта культурного-исторического наследия"""
    status: Optional[str] = None
    """Статус объекта туристского интереса"""
    potential: Optional[str] = None
    """Потенциальный интерес к объекту"""
    tourist: Optional[str] = None
    """Среднее количество туристов, посетивших объект туристского интереса за два предыдущих года(тыс. чел.)"""
    paid_services: Optional[str] = None
    """Средний объем платных услуг, оказанных объектом туристского интереса за два предыдущих года(млн. руб.)"""
    url: Optional[str] = None
    """Ссылка на портал, содержащий информацию об объекте туристского интереса"""


class Options39663(OptionProperties):
    title_key = "cad_number"
    type: Optional[str] = None
    """Вид объекта недвижимости"""
    registration_date: Optional[str] = None
    """Дата присвоения"""
    cad_number: Optional[str] = None
    """Кадастровый номер"""
    quarter_cad_number: Optional[str] = None
    """Кадастровый квартал"""
    readable_address: Optional[str] = None
    """Адрес"""
    name: Optional[str] = None
    """Наименование"""
    purpose: Optional[str] = None
    """Назначение"""
    common_data_status: Optional[str] = None
    """Статус"""
    cost_value: Optional[float] = None
    """Кадастровая стоимость (руб.)"""
    facility_cad_number: Optional[str] = None
    """Кадастровый номер предприятия как имущественного комплекса, если единый недвижимый комплекс (имущественное право на него) входит в состав предприятия как имущественного комплекса"""


class Options39664(OptionProperties):
    title_key = "name"
    type: Optional[str] = None
    """Вид объекта недвижимости"""
    registration_date: Optional[str] = None
    """Дата присвоения"""
    name: Optional[str] = None
    """Наименование ПИК при наличии такого наименования"""
    purpose: Optional[str] = None
    """Назначение ПИК"""
    cad_number: Optional[str] = None
    """Кадастровый номер"""
    quarter_cad_number: Optional[str] = None
    """Кадастровый квартал"""
    readable_address: Optional[str] = None
    """Адрес"""
    common_data_status: Optional[str] = None
    """Статус"""
    cost_value: Optional[str] = None
    """Стоимость предприятия (руб.)"""


class Options36991(OptionProperties):
    title_key = "loc"
    type_boundary_value: Optional[str] = None
    """Вид"""
    registration_date: Optional[str] = None
    """Дата присвоения"""
    brd_nmb: Optional[str] = None
    """Реестровый номер границы"""
    cadastral_district: Optional[str] = None
    """Кадастровый район"""
    old_account_number: Optional[str] = None
    """Учетный номер"""
    loc: Optional[str] = None
    """Наименование"""
    document_name: Optional[str] = None
    """Реквизиты решения об установлении или изменении границ"""
    legal_act_document_name: Optional[str] = None
    """Наименование документа"""
    legal_act_document_number: Optional[str] = None
    """Номер"""
    legal_act_document_date: Optional[str] = None
    """Дата"""
    legal_act_document_issuer: Optional[str] = None
    """Орган власти, организация выдавшие документ"""


class Layer872153Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options38967]]
):
    """Водная эрозия"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Водная эрозия",
            "layerTreeId": 33357,
            "layerId": 872153,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "Водная эрозия",
            "layerVisibleByDefault": False,
            "categoryId": 38967,
        }
    )


class Layer872155Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options38967]]
):
    """Линейная эрозия"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Линейная эрозия",
            "layerTreeId": 33358,
            "layerId": 872155,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "Линейная эрозия",
            "layerVisibleByDefault": False,
            "categoryId": 38967,
        }
    )


class Layer872164Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options38967]]
):
    """Ветровая эрозия"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Ветровая эрозия",
            "layerTreeId": 33359,
            "layerId": 872164,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "Ветровая эрозия",
            "layerVisibleByDefault": False,
            "categoryId": 38967,
        }
    )


class Layer872182Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options38967]]
):
    """Опустынивание"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Опустынивание",
            "layerTreeId": 33360,
            "layerId": 872182,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "Опустынивание",
            "layerVisibleByDefault": False,
            "categoryId": 38967,
        }
    )


class Layer872183Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options38967]]
):
    """Переувлажнение"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Переувлажнение",
            "layerTreeId": 33361,
            "layerId": 872183,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "Переувлажнение",
            "layerVisibleByDefault": False,
            "categoryId": 38967,
        }
    )


class Layer872202Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options38967]]
):
    """Подтопление"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Подтопление",
            "layerTreeId": 33362,
            "layerId": 872202,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "Подтопление",
            "layerVisibleByDefault": False,
            "categoryId": 38967,
        }
    )


class Layer872203Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options38967]]
):
    """Заболачивание"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Заболачивание",
            "layerTreeId": 33363,
            "layerId": 872203,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "Заболачивание",
            "layerVisibleByDefault": False,
            "categoryId": 38967,
        }
    )


class Layer872205Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options38967]]
):
    """Затопление"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Затопление",
            "layerTreeId": 33364,
            "layerId": 872205,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "Затопление",
            "layerVisibleByDefault": False,
            "categoryId": 38967,
        }
    )


class Layer872206Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options38967]]
):
    """Захламление"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Захламление",
            "layerTreeId": 33365,
            "layerId": 872206,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "Захламление",
            "layerVisibleByDefault": False,
            "categoryId": 38967,
        }
    )


class Layer872210Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options38967]]
):
    """Обвально-осыпные и оползневые процессы"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Обвально-осыпные и оползневые процессы",
            "layerTreeId": 33366,
            "layerId": 872210,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "Обвально-осыпные и оползневые процессы",
            "layerVisibleByDefault": False,
            "categoryId": 38967,
        }
    )


class Layer872211Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options38967]]
):
    """Абразия"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Абразия",
            "layerTreeId": 33367,
            "layerId": 872211,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "Абразия",
            "layerVisibleByDefault": False,
            "categoryId": 38967,
        }
    )


class Layer36945Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options38030]]
):
    """Кадастровые округа"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Кадастровые округа",
            "layerTreeId": 3,
            "layerId": 36945,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "ЕГРН. Кадастровые округа",
            "layerVisibleByDefault": False,
            "categoryId": 38030,
        }
    )


class Layer36316Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options36947]]
):
    """Территории объектов культурного наследия"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Территории объектов культурного наследия",
            "layerTreeId": 62,
            "layerId": 36316,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "Территории объекта культурного наследия (ЕГРН)",
            "layerVisibleByDefault": False,
            "categoryId": 36947,
        }
    )


class Layer36070Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options36382]]
):
    """Кадастровые районы"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Кадастровые районы ",
            "layerTreeId": 5873,
            "layerId": 36070,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "ЕГРН. Кадастровые районы",
            "layerVisibleByDefault": False,
            "categoryId": 36382,
        }
    )


class Layer872212Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options38967]]
):
    """Иные нарушенные земли"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Иные нарушенные земли ",
            "layerTreeId": 33368,
            "layerId": 872212,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "Иные нарушенные земли",
            "layerVisibleByDefault": False,
            "categoryId": 38967,
        }
    )


class Layer36071Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options36381]]
):
    """Кадастровые кварталы"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Кадастровые кварталы",
            "layerTreeId": 5,
            "layerId": 36071,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "ЕГРН. Кадастровые кварталы",
            "layerVisibleByDefault": False,
            "categoryId": 36381,
        }
    )


class Layer872213Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options38967]]
):
    """Нарушенные земли при наземном строительстве"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Нарушенные земли при наземном строительстве ",
            "layerTreeId": 33369,
            "layerId": 872213,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "Нарушенные земли при наземном строительстве",
            "layerVisibleByDefault": False,
            "categoryId": 38967,
        }
    )


class Layer872216Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options38967]]
):
    """Нарушенные земли при гидротехническом строительстве"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Нарушенные земли при гидротехническом строительстве",
            "layerTreeId": 33370,
            "layerId": 872216,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "Нарушенные земли при гидротехническом строительстве",
            "layerVisibleByDefault": False,
            "categoryId": 38967,
        }
    )


class Layer872217Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options38967]]
):
    """Нарушенные земли при недропользовании"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Нарушенные земли при недропользовании",
            "layerTreeId": 33371,
            "layerId": 872217,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "Нарушенные земли при недропользовании",
            "layerVisibleByDefault": False,
            "categoryId": 38967,
        }
    )


class Layer872218Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options38967]]
):
    """Нарушенные земли при промышленном лесопользовании"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Нарушенные земли при промышленном лесопользовании",
            "layerTreeId": 33372,
            "layerId": 872218,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "Нарушенные земли при промышленном лесопользовании",
            "layerVisibleByDefault": False,
            "categoryId": 38967,
        }
    )


class Layer872219Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options38967]]
):
    """Нарушенные земли при сельскохозяйственном освоении"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Нарушенные земли при сельскохозяйственном освоении",
            "layerTreeId": 33373,
            "layerId": 872219,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "Нарушенные земли при сельскохозяйственном освоении",
            "layerVisibleByDefault": False,
            "categoryId": 38967,
        }
    )


class Layer872220Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options38967]]
):
    """Нарушенные земли при проведении геологоразведочных, испытательных, эксплуатационных и иных работ"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Нарушенные земли при проведении геологоразведочных, испытательных, эксплуатационных и иных работ",
            "layerTreeId": 33374,
            "layerId": 872220,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "Нарушенные земли при проведении геологоразведочных, испытательных, эксплуатационных и иных работ",
            "layerVisibleByDefault": False,
            "categoryId": 38967,
        }
    )


class Layer872221Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options38967]]
):
    """Нарушенные земли при складировании и захоронении промышленных отходов, загрязнение земель"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Нарушенные земли при складировании и захоронении промышленных отходов, загрязнение земель",
            "layerTreeId": 33375,
            "layerId": 872221,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "Нарушенные земли при складировании и захоронении промышленных отходов, загрязнение земель",
            "layerVisibleByDefault": False,
            "categoryId": 38967,
        }
    )


class Layer37313Feature(
    BaseFeature[Union[MultiLineString, LineString], NspdProperties[Options38998]]
):
    """Государственная граница Российской Федерации"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Государственная граница Российской Федерации",
            "layerTreeId": 7,
            "layerId": 37313,
            "layerType": "wms",
            "geometryType": "LINESTRING",
            "layerName": "Государственная граница Российской Федерации",
            "layerVisibleByDefault": False,
            "categoryId": 38998,
        }
    )


class Layer872222Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options38967]]
):
    """Гари"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Гари",
            "layerTreeId": 33376,
            "layerId": 872222,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "Гари",
            "layerVisibleByDefault": False,
            "categoryId": 38967,
        }
    )


class Layer872224Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options38967]]
):
    """Негативный процесс отсутствует"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Негативный процесс отсутствует",
            "layerTreeId": 33377,
            "layerId": 872224,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "Негативный процесс отсутствует",
            "layerVisibleByDefault": False,
            "categoryId": 38967,
        }
    )


class Layer37314Feature(
    BaseFeature[Union[MultiLineString, LineString], NspdProperties[Options38999]]
):
    """Субъекты Российской Федерации (линии)"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Субъекты Российской Федерации (линии)",
            "layerTreeId": 9,
            "layerId": 37314,
            "layerType": "wms",
            "geometryType": "LINESTRING",
            "layerName": "Границы между субъектами Российской Федерации (линии)",
            "layerVisibleByDefault": False,
            "categoryId": 38999,
        }
    )


class Layer37315Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options39000]]
):
    """Субъекты Российской Федерации (полигоны)"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Субъекты Российской Федерации (полигоны)",
            "layerTreeId": 10,
            "layerId": 37315,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "Границы между субъектами Российской Федерации (полигоны)",
            "layerVisibleByDefault": False,
            "categoryId": 39000,
        }
    )


class Layer872262Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options38967]]
):
    """Засоление"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Засоление",
            "layerTreeId": 33378,
            "layerId": 872262,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "Засоление",
            "layerVisibleByDefault": False,
            "categoryId": 38967,
        }
    )


class Layer36278Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options36829]]
):
    """Муниципальные образования (полигональный)"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Муниципальные образования (полигональный)",
            "layerTreeId": 12,
            "layerId": 36278,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "Муниципальные образования (полигональный)",
            "layerVisibleByDefault": False,
            "categoryId": 36829,
        }
    )


class Layer36279Feature(
    BaseFeature[Union[MultiLineString, LineString], NspdProperties[Options36830]]
):
    """Муниципальные образования (линейный)"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Муниципальные образования (линейный)",
            "layerTreeId": 13,
            "layerId": 36279,
            "layerType": "wms",
            "geometryType": "LINESTRING",
            "layerName": "Муниципальные образования (линейный)",
            "layerVisibleByDefault": False,
            "categoryId": 36830,
        }
    )


class Layer36281Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options36832]]
):
    """Населённые пункты (полигоны)"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Населённые пункты (полигоны)",
            "layerTreeId": 15,
            "layerId": 36281,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "Границы населенных пунктов",
            "layerVisibleByDefault": False,
            "categoryId": 36832,
        }
    )


class Layer37316Feature(
    BaseFeature[Union[MultiLineString, LineString], NspdProperties[Options38996]]
):
    """Населённые пункты (линии)"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Населённые пункты (линии)",
            "layerTreeId": 16,
            "layerId": 37316,
            "layerType": "wms",
            "geometryType": "LINESTRING",
            "layerName": "Границы населенных пунктов (линии)",
            "layerVisibleByDefault": False,
            "categoryId": 38996,
        }
    )


class Layer36048Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options36368]]
):
    """Земельные участки из ЕГРН"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Земельные участки из ЕГРН",
            "layerTreeId": 72,
            "layerId": 36048,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "Росреестр: Земельные участки ЕГРН",
            "layerVisibleByDefault": False,
            "categoryId": 36368,
        }
    )


class Layer37294Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options38943]]
):
    """Земельные участки, образуемые по схеме расположения земельного участка"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Земельные участки, образуемые по схеме расположения земельного участка",
            "layerTreeId": 22,
            "layerId": 37294,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "Схемы расположения земельных участков",
            "layerVisibleByDefault": False,
            "categoryId": 38943,
        }
    )


class Layer37299Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options38981]]
):
    """Земельные участки, выставленные на аукцион"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Земельные участки, выставленные на аукцион ",
            "layerTreeId": 23,
            "layerId": 37299,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "Земельные участки, выставленные на аукцион",
            "layerVisibleByDefault": False,
            "categoryId": 38981,
        }
    )


class Layer37298Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options38979]]
):
    """Земельные участки, свободные от прав третьих лиц"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Земельные участки, свободные от прав третьих лиц",
            "layerTreeId": 24,
            "layerId": 37298,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "Земельные участки свободные от прав третьих лиц",
            "layerVisibleByDefault": False,
            "categoryId": 38979,
        }
    )


class Layer849407Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options445347]]
):
    """Земля для стройки ПКК"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Земля для стройки ПКК",
            "layerTreeId": 6409,
            "layerId": 849407,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "Земля для стройки",
            "layerVisibleByDefault": False,
            "categoryId": 445347,
        }
    )


class Layer849453Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options445397]]
):
    """Земля для туризма ПКК"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Земля для туризма ПКК",
            "layerTreeId": 6410,
            "layerId": 849453,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "Земля для туризма",
            "layerVisibleByDefault": False,
            "categoryId": 445397,
        }
    )


class Layer36473Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options37158]]
):
    """Земельные участки, образуемые по проекту межевания территории"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Земельные участки, образуемые по проекту межевания территории",
            "layerTreeId": 21,
            "layerId": 36473,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "Земельные участки, образуемые по проекту межевания территории (полигональный)",
            "layerVisibleByDefault": False,
            "categoryId": 37158,
        }
    )


class Layer36049Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options36369]]
):
    """Здания"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Здания",
            "layerTreeId": 29,
            "layerId": 36049,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "Росреестр: Здания ЕГРН",
            "layerVisibleByDefault": False,
            "categoryId": 36369,
        }
    )


class Layer36328Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options36383]]
):
    """Сооружения"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Сооружения",
            "layerTreeId": 30,
            "layerId": 36328,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "Росреестр: Сооружения ЕГРН",
            "layerVisibleByDefault": False,
            "categoryId": 36383,
        }
    )


class Layer36329Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options36384]]
):
    """Объекты незавершенного строительства"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Объекты незавершенного строительства",
            "layerTreeId": 31,
            "layerId": 36329,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "Росреестр: Объекты незавершённого строительства ЕГРН",
            "layerVisibleByDefault": False,
            "categoryId": 36384,
        }
    )


class Layer37433Feature(
    BaseFeature[Union[MultiPoint, Point], NspdProperties[Options39663]]
):
    """Единые недвижимые комплексы"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Единые недвижимые комплексы",
            "layerTreeId": 35,
            "layerId": 37433,
            "layerType": "wms",
            "geometryType": "POINT",
            "layerName": "Единые недвижимые комплексы",
            "layerVisibleByDefault": False,
            "categoryId": 39663,
        }
    )


class Layer37434Feature(
    BaseFeature[Union[MultiPoint, Point], NspdProperties[Options39664]]
):
    """Предприятие как имущественный комплекс"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Предприятие как имущественный комплекс",
            "layerTreeId": 36,
            "layerId": 37434,
            "layerType": "wms",
            "geometryType": "POINT",
            "layerName": "Предприятие как имущественный комплекс",
            "layerVisibleByDefault": False,
            "categoryId": 39664,
        }
    )


class Layer37577Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options36940]]
):
    """ЗОУИТ объектов культурного наследия"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "ЗОУИТ объектов культурного наследия",
            "layerTreeId": 39,
            "layerId": 37577,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "ЗОУИТ объектов культурного наследия",
            "layerVisibleByDefault": False,
            "categoryId": 36940,
        }
    )


class Layer37578Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options36940]]
):
    """ЗОУИТ объектов энергетики, связи, транспорта"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "ЗОУИТ объектов энергетики, связи, транспорта",
            "layerTreeId": 40,
            "layerId": 37578,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "ЗОУИТ объектов энергетики, связи, транспорта",
            "layerVisibleByDefault": False,
            "categoryId": 36940,
        }
    )


class Layer37580Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options36940]]
):
    """ЗОУИТ природных территорий"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "ЗОУИТ природных территорий",
            "layerTreeId": 41,
            "layerId": 37580,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "ЗОУИТ природных территорий",
            "layerVisibleByDefault": False,
            "categoryId": 36940,
        }
    )


class Layer37579Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options36940]]
):
    """ЗОУИТ охраняемых объектов и безопасности"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "ЗОУИТ охраняемых объектов и безопасности",
            "layerTreeId": 42,
            "layerId": 37579,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "ЗОУИТ охраняемых объектов и безопасности",
            "layerVisibleByDefault": False,
            "categoryId": 36940,
        }
    )


class Layer37581Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options36940]]
):
    """Иные ЗОУИТ"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Иные ЗОУИТ",
            "layerTreeId": 43,
            "layerId": 37581,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "Иные ЗОУИТ",
            "layerVisibleByDefault": False,
            "categoryId": 36940,
        }
    )


class Layer36315Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options36946]]
):
    """Территориальные зоны"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Территориальные зоны",
            "layerTreeId": 45,
            "layerId": 36315,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "Территориальные зоны (ЕГРН)",
            "layerVisibleByDefault": False,
            "categoryId": 36946,
        }
    )


class Layer37293Feature(
    BaseFeature[Union[MultiLineString, LineString], NspdProperties[Options38942]]
):
    """Красные линии"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Красные линии ",
            "layerTreeId": 46,
            "layerId": 37293,
            "layerType": "wms",
            "geometryType": "LINESTRING",
            "layerName": "Красные линии",
            "layerVisibleByDefault": False,
            "categoryId": 38942,
        }
    )


class Layer36317Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options36948]]
):
    """Особо охраняемые природные территории"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Особо охраняемые природные территории ",
            "layerTreeId": 49,
            "layerId": 36317,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "Особо охраняемая природные территории (ЕГРН)",
            "layerVisibleByDefault": False,
            "categoryId": 36948,
        }
    )


class Layer36311Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options36942]]
):
    """Охотничьи угодья"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Охотничьи угодья ",
            "layerTreeId": 50,
            "layerId": 36311,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "Охотничьи угодья (ЕГРН)",
            "layerVisibleByDefault": False,
            "categoryId": 36942,
        }
    )


class Layer36314Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options36945]]
):
    """Лесничества"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Лесничества",
            "layerTreeId": 51,
            "layerId": 36314,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "Лесничества (ЕГРН)",
            "layerVisibleByDefault": False,
            "categoryId": 36945,
        }
    )


class Layer843763Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options278838]]
):
    """Граница лесопарка"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Граница лесопарка",
            "layerTreeId": 22395,
            "layerId": 843763,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "Граница лесопарка (ЕГРН)",
            "layerVisibleByDefault": False,
            "categoryId": 278838,
        }
    )


class Layer36469Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options36992]]
):
    """Береговые линии (границы водных объектов) (полигональный)"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Береговые линии (границы водных объектов) (полигональный)",
            "layerTreeId": 55,
            "layerId": 36469,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "Сведения о береговых линиях (границах водных объектов). (полигональный)",
            "layerVisibleByDefault": False,
            "categoryId": 36992,
        }
    )


class Layer36470Feature(
    BaseFeature[Union[MultiLineString, LineString], NspdProperties[Options37148]]
):
    """Береговые линии (границы водных объектов)(линейный)"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Береговые линии (границы водных объектов)(линейный)",
            "layerTreeId": 56,
            "layerId": 36470,
            "layerType": "wms",
            "geometryType": "LINESTRING",
            "layerName": "Сведения о береговых линиях (границах водных объектов). (линейный)",
            "layerVisibleByDefault": False,
            "categoryId": 37148,
        }
    )


class Layer36303Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options36941]]
):
    """Особые экономические зоны"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Особые экономические зоны",
            "layerTreeId": 58,
            "layerId": 36303,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "Особые экономические зоны (ЕГРН)",
            "layerVisibleByDefault": False,
            "categoryId": 36941,
        }
    )


class Layer36312Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options36943]]
):
    """Территории опережающего развития"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Территории опережающего развития",
            "layerTreeId": 59,
            "layerId": 36312,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "Территории опережающего развития (ЕГРН)",
            "layerVisibleByDefault": False,
            "categoryId": 36943,
        }
    )


class Layer36471Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options36991]]
):
    """Игорные зоны"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Игорные зоны",
            "layerTreeId": 61,
            "layerId": 36471,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "Границы игорных зон",
            "layerVisibleByDefault": False,
            "categoryId": 36991,
        }
    )


class Layer37430Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options39228]]
):
    """Территории выполнения комплексных кадастровых работ"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Территории выполнения комплексных кадастровых работ",
            "layerTreeId": 66,
            "layerId": 37430,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "ЕГРН.Территории выполнения комплексных кадастровых работ",
            "layerVisibleByDefault": False,
            "categoryId": 39228,
        }
    )


class Layer37295Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options38944]]
):
    """Территория проведения мероприятий по ликвидации накопленного вреда окружающей среде, образовавшегося в результате производства химической продукции в г. Усолье-Сибирское Иркутской области"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Территория проведения мероприятий по ликвидации накопленного вреда окружающей среде, образовавшегося в результате производства химической продукции в г. Усолье-Сибирское Иркутской области",
            "layerTreeId": 64,
            "layerId": 37295,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "Территория проведения мероприятий по ликвидации накопленного вреда окружающей среде, образовавшегося в результате производства химической продукции в г. Усолье-Сибирское Иркутской области",
            "layerVisibleByDefault": False,
            "categoryId": 38944,
        }
    )


class Layer849601Feature(
    BaseFeature[Union[MultiPoint, Point], NspdProperties[Options445547]]
):
    """Объекты туристского интереса"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Объекты туристского интереса",
            "layerTreeId": 6487,
            "layerId": 849601,
            "layerType": "wms",
            "geometryType": "POINT",
            "layerName": "Объекты туристского интереса",
            "layerVisibleByDefault": False,
            "categoryId": 445547,
        }
    )


class Layer37236Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options36368]]
):
    """Кадастровая стоимость объекта"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Кадастровая стоимость объекта",
            "layerTreeId": 69,
            "layerId": 37236,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "Росреестр: Земельные участки ЕГРН. Кадастровая стоимость Тепловая карта",
            "layerVisibleByDefault": False,
            "categoryId": 36368,
        }
    )


class Layer37758Feature(
    BaseFeature[Union[MultiPolygon, Polygon, Point], NspdProperties[Options36368]]
):
    """Удельный показатель кадастровой стоимости"""

    layer_meta = LayerNode.model_validate(
        {
            "title": "Удельный показатель кадастровой стоимости",
            "layerTreeId": 70,
            "layerId": 37758,
            "layerType": "wms",
            "geometryType": "POLYGON",
            "layerName": "Росреестр: Земельные участки ЕГРН. Удельный Показатель Кадастровой Стоимости Тепловая карта",
            "layerVisibleByDefault": False,
            "categoryId": 36368,
        }
    )



from typing import Annotated, Optional, Union

from geojson_pydantic import FeatureCollection
from pydantic import BaseModel, BeforeValidator, RootModel

from pynspd.schemas import NspdFeature
from pynspd.schemas._common import CamelModel


class BadResponse(CamelModel):
    """Ошибка на поисковой запрос"""

    code: int
    message: str
    request_id: str


class SearchResponse(CamelModel):
    """Положительный ответ на поисковой запрос"""

    data: FeatureCollection[NspdFeature]
    meta: list["Meta"]


class Meta(CamelModel):
    """Метаданные ответа на поисковой запрос"""

    total_count: int
    category_id: int


ResponseModel = RootModel[Union[SearchResponse, BadResponse]]


class NspdTabResponse(BaseModel):
    title: str
    value: Annotated[
        Optional[list[str]],
        BeforeValidator(
            lambda x: None if isinstance(x, list) and len(x) == 1 and x[0] == "" else x
        ),
    ]


class NspdTabGroupResponse(BaseModel):
    title: str
    object: list[NspdTabResponse]



import httpx


class PynspdError(Exception):
    """Базовый класс ошибок библиотеки"""


class UnknownLayer(PynspdError):
    """Не найдено определение слоя (по имени, категории и тд)"""


class AmbiguousSearchError(PynspdError):
    """Поисковой запрос выдал неоднозначный результат"""

    def __init__(self, query: str):
        msg = f"Найдено больше одного объекта по запросу {query}"
        super().__init__(msg)


class TooBigContour(PynspdError):
    """Контур слишком большой, сервер не может ответить"""

    def __init__(self):
        super().__init__(
            "Попробуйте уменьшить площадь поиска "
            "или воспользоваться методом `.search_in_contour_iter(...)`",
        )


class PynspdResponseError(PynspdError):
    """Базовый класс ошибок для неуспешных запросов"""

    def __init__(self, response: httpx.Response):
        self.response = response
        try:
            details: str = response.json()["message"]
        except (KeyError, ValueError):
            details = "-"
        msg = (
            f"{self.__class__.__doc__}\n"
            f"Status code: {response.status_code}\n"
            f"Details: {details}"
        )
        super().__init__(msg)


class PynspdServerError(PynspdResponseError):
    """Серверная ошибка"""


class BlockedIP(PynspdResponseError):
    """Доступ заблокирован"""


class TooManyRequests(PynspdResponseError):
    """Слишком много запросов"""


class NotFound(PynspdResponseError):
    """Ничего не найдено"""



from pynspd import Nspd


def test_more_than_one_request_filter(api: Nspd):
    """Когда на поисковой запрос выпадает больше одного объекта"""
    feat = api.find("77:04:0004018:3371")
    assert feat is not None


def test_iter_cn(api: Nspd):
    assert list(api.iter_cn("77:04:0004018:3371; 77:04:0004018:3372")) == [
        "77:04:0004018:3371",
        "77:04:0004018:3372",
    ]



import json
import re
import warnings
from functools import wraps
from hashlib import md5
from pathlib import Path
from time import sleep
from typing import Any, Generator, Literal, Optional, Type, Union

import mercantile
import numpy as np
import ua_generator
from hishel import (
    BaseStorage,
    CacheTransport,
    FileStorage,
    RedisStorage,
    SQLiteStorage,
)
from httpx import (
    BaseTransport,
    Client,
    ConnectError,
    HTTPError,
    HTTPTransport,
    ProxyError,
    RemoteProtocolError,
    Response,
    TimeoutException,
)
from httpx._types import ProxyTypes, QueryParamTypes
from shapely import MultiPolygon, Point, Polygon, box, to_geojson

try:
    import sqlite3
except ImportError:
    sqlite3 = None  # type: ignore
try:
    import redis
except ImportError:
    redis = None  # type: ignore

import pynspd.errors as err
from pynspd.client import (
    NSPD_CACHE_CONTROLLER,
    SSL_CONTEXT,
    BaseNspdClient,
)
from pynspd.logger import logger
from pynspd.map_types.enums import TabTitle, ThemeId
from pynspd.schemas import Layer36048Feature, Layer36049Feature, NspdFeature
from pynspd.schemas.feature import Feat
from pynspd.schemas.responses import (
    NspdTabGroupResponse,
    NspdTabResponse,
    SearchResponse,
)


def retry_on_http_error(func):
    """Декоратор для повторения запроса при ошибках запроса"""

    @wraps(func)
    def wrapper(self: "Nspd", *args, **kwargs):
        attempt = -1
        last_error: Optional[Exception] = None
        while True:
            attempt += 1
            logger_suffix = f'Wrapped method "{func.__name__}", retry {attempt} -'
            if attempt > self._retries:
                logger.debug("%s run out attempts", logger_suffix)
                if last_error is None:
                    last_error = Exception(
                        f"Too many retries ({attempt}/{self._retries})"
                    )
                raise last_error
            try:
                logger.debug("%s start request", logger_suffix)
                return func(self, *args, **kwargs)
            except HTTPError as e:
                last_error = e
                if isinstance(e, RemoteProtocolError):
                    logger.debug("%s remote protocol error", logger_suffix)
                elif isinstance(e, ConnectError):
                    logger.debug("%s connection error", logger_suffix)
                elif isinstance(e, TimeoutException):
                    logger.debug("%s timeout", logger_suffix)
                    sleep(1)
                else:
                    logger.exception("%s unexpected exception", logger_suffix)
                    raise e
            except err.PynspdServerError as e:
                logger.debug("%s unknown server error", logger_suffix)
                last_error = e
            except err.BlockedIP as e:
                logger.debug("%s blocked IP, retrying", logger_suffix)
                last_error = e
                if not self._retry_on_blocked_ip:
                    raise e
            except err.TooManyRequests as e:
                logger.debug("%s too many requests", logger_suffix)
                last_error = e
                sleep(1)

    return wrapper


class Nspd(BaseNspdClient):
    """Клиент для НСПД

    Example:
    ```python
    with pynspd.Nspd() as nspd:
        feat = nspd.find_zu("77:05:0001005:19")
    ```

    Args:
        client_timeout:
            Время ожидания ответа.
            Если не установлен - есть вероятность бесконечного ожидания. По умолчанию `None`.
        client_retries:
            Количество попыток при неудачном запросе
            (таймаут, неожиданный обрыв соединения, 5хх ошибки). По умолчанию `10`.
        client_retry_on_blocked_ip:
            При получении ошибки 403 (доступ заблокирован для вашего IP),
            продолжать попытки запроса до исчерпания retries.
            Рекомендуется использовать только c ротируемыми прокси. По умолчанию `False`.
        client_proxy:
            Адрес для проксирования запросов. По умолчанию `None`.
        client_dns_resolve:
            Использовать в запросах IP адрес НСПД вместо доменного имени.
            Рекомендуется включить, если используемый прокси
            не может сам разрешать доменные имена.
            По умолчанию `False`.
        cache_folder_path:
            Путь до папки для кэша запросов.. По умолчанию `None`.
        cache_sqlite_url:
            Строка подключения для sqlite-хранилища кэша. По умолчанию `None`.
        cache_redis_url:
            Строка подключения для redis-хранилища кэша. По умолчанию `None`.
        cache_ttl:
            Количество времени (в секундах) сколько будет храниться кэш. По умолчанию `None`.
        cache_storage:
            Ручная настройка объекта хранилища кэша (см. https://hishel.com/advanced/storages/). По умолчанию `None`.
        trust_env:
            Использовать переменные окружения для инициализации.
            По умолчанию `True`.
    """

    def __init__(
        self,
        *,
        client_timeout: Optional[int] = None,
        client_retries: Optional[int] = None,
        client_retry_on_blocked_ip: Optional[bool] = None,
        client_proxy: Optional[ProxyTypes] = None,
        client_dns_resolve: Optional[bool] = None,
        cache_storage: Optional[BaseStorage] = None,
        cache_folder_path: Optional[Union[str, Path]] = None,
        cache_sqlite_url: Optional[str] = None,
        cache_redis_url: Optional[str] = None,
        cache_ttl: Optional[int] = None,
        trust_env: bool = True,
    ):
        self._timeout = self._int_var("client_timeout", client_timeout, trust_env)
        self._retries = self._int_var("client_retries", client_retries, trust_env)
        if self._retries is None:
            self._retries = 10
        self._retry_on_blocked_ip = (
            self._bool_var(
                "client_retry_on_blocked_ip", client_retry_on_blocked_ip, trust_env
            )
            or False
        )
        self._proxy = self._str_var("client_proxy", client_proxy, trust_env)
        self._dns_resolve = (
            self._bool_var("client_dns_resolve", client_dns_resolve, trust_env) or False
        )

        self._cache_folder_path = self._str_var(
            "cache_folder_path", cache_folder_path, trust_env
        )
        self._cache_sqlite_url = self._str_var(
            "cache_sqlite_url", cache_sqlite_url, trust_env
        )
        self._cache_redis_url = self._str_var(
            "cache_redis_url", cache_redis_url, trust_env
        )
        self._cache_ttl = self._int_var("cache_ttl", cache_ttl, trust_env)
        self._cache_storage = cache_storage
        if (
            sum(
                [
                    int(bool(i))
                    for i in (
                        self._cache_folder_path,
                        self._cache_sqlite_url,
                        self._cache_redis_url,
                        self._cache_ttl,
                        self._cache_storage,
                    )
                ]
            )
            > 1
        ):
            raise ValueError("Допустимо выбрать только один вариант хранилища кэша")

        self._client: Optional[Client] = None
        self._last_response: Optional[Response] = None

    def _build_cache_storage(self) -> Optional[BaseStorage]:
        if self._cache_folder_path is not None:
            return FileStorage(
                base_path=Path(self._cache_folder_path), ttl=self._cache_ttl
            )
        elif self._cache_sqlite_url is not None:
            if sqlite3 is None:
                raise RuntimeError(
                    "Не установлены необходимые модули для кэширования этого типа. "
                    "Убедитесь, что вы установили `pynspd` с расширением `sqlite`.\n"
                    "```pip install pynspd[sqlite]```"
                )
            conn = sqlite3.connect(self._cache_sqlite_url, check_same_thread=False)
            return SQLiteStorage(connection=conn, ttl=self._cache_ttl)
        elif self._cache_redis_url is not None:
            if redis is None:
                raise RuntimeError(
                    "Не установлены необходимые модули для кэширования этого типа. "
                    "Убедитесь, что вы установили `pynspd` с расширением `redis`.\n"
                    "```pip install pynspd[redis]```"
                )
            client = redis.Redis.from_url(self._cache_redis_url)
            return RedisStorage(client=client, ttl=self._cache_ttl)
        return None

    def _build_client(self) -> Client:
        transport: BaseTransport = HTTPTransport(
            verify=SSL_CONTEXT, retries=self._retries, proxy=self._proxy
        )
        if self._cache_storage is None:
            self._cache_storage = self._build_cache_storage()
        if self._cache_storage is not None:
            transport = CacheTransport(
                transport=transport,
                storage=self._cache_storage,
                controller=NSPD_CACHE_CONTROLLER,
            )

        host = self.DNS_HOST if not self._dns_resolve else self.IP_HOST
        return Client(
            base_url="https://" + host,
            timeout=self._timeout,
            headers={
                "User-Agent": ua_generator.generate().text,
                "Referer": self.DNS_URL,
                "Host": self.DNS_HOST,
            },
            transport=transport,
        )

    def _rebuild_client_with_dns_resolve(self):
        assert not self._dns_resolve
        assert self._client is not None
        assert self._client.base_url == self.DNS_URL
        warnings.warn(
            "Текущий прокси не может сам разрешать доменные имена, "
            "рекомендуется использовать client_dns_resolve=True, "
            "чтобы клиент заранее подставлял IP-адреса и уменьшал нагрузку на соединение",
            stacklevel=7,
        )
        self._client.close()
        self._dns_resolve = True
        self._client = self._build_client()

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.close()

    def close(self):
        """Завершение сессии"""
        if self._client is not None:
            self._client.close()

    def request(
        self,
        method: str,
        url: str,
        params: Optional[QueryParamTypes] = None,
        json: Optional[dict] = None,
    ) -> Response:
        """Базовый запрос к API НСПД"""
        logger.debug("Request %s", url)
        if self._client is None:
            self._client = self._build_client()
        try:
            r = self._client.request(method, url, params=params, json=json)
        except ProxyError as e:
            # Прокси не поддерживает DNS-resolve, необходимо использовать IP-адрес
            # https://github.com/encode/httpx/issues/203#issuecomment-1017726203
            msg = str(e)
            if (
                "Connection not allowed by ruleset" in msg
                or "503 Target host denied" in msg
            ) and not self._dns_resolve:
                logger.debug("Proxy can't resolve dns; change to ip mode")
                self._rebuild_client_with_dns_resolve()
                return self.request(method, url, params, json)
            raise e
        self._last_response = r
        code = r.status_code
        if code == 403:
            raise err.BlockedIP(r)
        if code == 404:
            raise err.NotFound(r)
        if code == 429:
            raise err.TooManyRequests(r)
        if code >= 500:
            raise err.PynspdServerError(r)
        if not str(code).startswith("2"):
            raise err.PynspdResponseError(r)
        return r

    @retry_on_http_error
    def safe_request(
        self,
        method: str,
        url: str,
        params: Optional[QueryParamTypes] = None,
        json: Optional[dict] = None,
    ) -> Response:
        """Базовый запрос к api НСПД с обработкой ошибок"""
        return self.request(method, url, params, json)

    ####################
    ### QUERY SEARCH ###
    ####################

    @retry_on_http_error
    def _search(self, params: dict[str, Any]) -> Optional[list[NspdFeature]]:
        """Базовый поисковый запрос на НСПД"""
        try:
            r = self.request("get", "/api/geoportal/v2/search/geoportal", params=params)
            return SearchResponse.model_validate(r.json()).data.features
        except err.NotFound:
            return None

    def search(
        self, query: str, theme_id: ThemeId = ThemeId.REAL_ESTATE_OBJECTS
    ) -> Optional[list[NspdFeature]]:
        """Поисковой запрос по предустановленной теме

        Args:
            query: Поисковой запрос
            theme_id:
                Вид объекта (кадастровое деление, объект недвижимости и т.д.).
                По умолчанию: объекты недвижимости

        Returns:
            Положительный ответ от сервиса, либо None, если ничего не найдено
        """
        return self._search(
            params={
                "query": query,
                "thematicSearchId": theme_id.value,
            }
        )

    def search_in_layer(
        self, query: str, layer_def: Type[Feat]
    ) -> Optional[list[Feat]]:
        """Поиск по определению слоя

        Args:
            query: Поисковой запрос
            layer_def: Определение слоя

        Returns:
            Массив валидированных объектов, если что-то найдено
        """
        raw_features = self._search(
            params={
                "query": query,
                "layersId": layer_def.layer_meta.layer_id,
            }
        )
        return self._cast_features_to_layer_defs(raw_features, layer_def)

    def search_in_layers(
        self, query: str, *layer_defs: Type[Feat]
    ) -> Optional[list[NspdFeature]]:
        """Поиск по определениям слоев

        Args:
            query: Поисковой запрос
            layer_defs: Определения слоев

        Returns:
            Массив невалидированных объектов, если что-то найдено
        """
        return self._search(
            params={
                "query": query,
                "layersId": [i.layer_meta.layer_id for i in layer_defs],
            }
        )

    def find(
        self, query: str, theme_id: ThemeId = ThemeId.REAL_ESTATE_OBJECTS
    ) -> Optional[NspdFeature]:
        """Найти объект по предустановленной теме

        Args:
            query: Поисковой запрос
            theme_id:
                Вид объекта (кадастровое деление, объект недвижимости и т.д.).
                По умолчанию: объекты недвижимости

        Returns:
            Положительный ответ от сервиса, либо None, если ничего не найдено
        """
        return self._filter_search_by_query(self.search(query, theme_id), query)

    def find_in_layer(self, query: str, layer_def: Type[Feat]) -> Optional[Feat]:
        """Найти объект по определению слоя

        Args:
            query: Поисковой запрос
            layer_def: Определение слоя

        Returns:
            Валидированная модель слоя, если найдено
        """
        return self._filter_search_by_query(
            self.search_in_layer(query, layer_def), query
        )

    ######################
    ### POLYGON SEARCH ###
    ######################

    @retry_on_http_error
    def _search_in_contour(
        self,
        countour: Union[Polygon, MultiPolygon],
        *category_ids: int,
    ) -> Optional[list[NspdFeature]]:
        """Поиск объектов в контуре по ID категорий слоев

        Args:
            countour: Геометрический объект с контуром
            category_ids: ID категорий слоев

        Returns:
            Список объектов, пересекающихся с контуром, если найден хоть один
        """
        feature_geojson = json.loads(to_geojson(countour))
        feature_geojson["crs"] = {
            "type": "name",
            "properties": {"name": "EPSG:4326"},
        }
        payload = {
            "categories": [{"id": id_} for id_ in category_ids],
            "geom": {
                "type": "FeatureCollection",
                "features": [
                    {"geometry": feature_geojson, "type": "Feature", "properties": {}}
                ],
            },
        }
        try:
            response = self.request(
                "post",
                "/api/geoportal/v1/intersects",
                params={"typeIntersect": "fullObject"},
                json=payload,
            )
            return self._validate_feature_collection_response(response)
        except err.PynspdServerError as e:
            if '"code":400004' in e.response.text:
                raise err.TooBigContour from e
            raise e

    def search_in_contour(
        self,
        countour: Union[Polygon, MultiPolygon],
        layer_def: Type[Feat],
    ) -> Optional[list[Feat]]:
        """Поиск объектов слоя в контуре

        Args:
            countour: Геометрический объект с контуром
            layer_def: Модель слоя

        Raises:
            TooBigContour: НСПД не может обработать такой контур

        Returns:
            Список объектов, пересекающихся с контуром, если найден хоть один
        """
        raw_features = self._search_in_contour(
            countour, layer_def.layer_meta.category_id
        )
        return self._cast_features_to_layer_defs(raw_features, layer_def)

    def _iter_search_in_box(
        self,
        xmin: float,
        ymin: float,
        xmax: float,
        ymax: float,
        layer_def: Type[Feat],
    ) -> Generator[Feat, None, None]:
        """Рекурсивный поиск объектов в границах"""

        def split_extent(xmin: float, ymin: float, xmax: float, ymax: float):
            midx = (xmax + xmin) / 2
            midy = (ymax + ymin) / 2
            yield xmin, ymin, midx, midy
            yield midx, midy, xmax, ymax
            yield midx, ymin, xmax, midy
            yield xmin, midy, midx, ymax

        try:
            logger_prefix = "Search [%.2f, %.2f, %.2f, %.2f]: " % (
                xmin,
                ymin,
                xmax,
                ymax,
            )
            logger.debug(logger_prefix + "start")
            feats = self.search_in_contour(box(xmin, ymin, xmax, ymax), layer_def)
            if feats is None:
                logger.debug(logger_prefix + "empty")
                return
            logger.debug(logger_prefix + "success")
            for f in feats:
                yield f
        except err.TooBigContour:
            logger.debug(logger_prefix + "failed, split tiles")
            for sp_xmin, sp_ymin, sp_xmax, sp_ymax in split_extent(
                xmin, ymin, xmax, ymax
            ):
                for f in self._iter_search_in_box(
                    sp_xmin, sp_ymin, sp_xmax, sp_ymax, layer_def
                ):
                    yield f

    def search_in_contour_iter(
        self,
        countour: Union[Polygon, MultiPolygon],
        layer_def: Type[Feat],
        *,
        only_intersects: bool = False,
    ) -> Generator[Feat, None, None]:
        """Поиск объектов в указанных границах.

        Внимание: количество запросов кратно зависит от площади поиска.
        Если вы хотите вручную обрабатывать ошибку `TooBigContour`,
        используйте метод `search_in_contour(...)`

        Args:
            countour: Геометрический объект с контуром
            layer_def: Модель слоя
            only_intersects:
                Возвращать только те объекты,
                которые пересекаются с изначальным контуром. По умолчанию False

        Returns:
            Генератор объектов слоя в указанной области
        """
        cache = set()
        xmin, ymin, xmax, ymax = countour.bounds
        for feat in self._iter_search_in_box(xmin, ymin, xmax, ymax, layer_def):
            cache_id = md5(feat.model_dump_json().encode()).hexdigest()
            if cache_id in cache:
                continue
            cache.add(cache_id)
            if only_intersects and not countour.intersects(feat.geometry.to_shape()):
                continue
            yield feat

    ####################
    ### POINT SEARCH ###
    ####################

    @retry_on_http_error
    def _search_at_point(self, pt: Point, layer_id: int) -> Optional[list[NspdFeature]]:
        """Поиск объектов слоя в точке"""
        tile_size = 512
        tile = mercantile.tile(
            pt.x, pt.y, zoom=24
        )  # zoom=24 должно быть достаточно для самого точного совпадения
        tile_bounds = mercantile.bounds(tile)
        i = np.interp(pt.x, [tile_bounds.west, tile_bounds.east], [0, tile_size])
        j = np.interp(pt.y, [tile_bounds.south, tile_bounds.north], [0, tile_size])
        bbox = ",".join(map(str, tile_bounds))
        params = {
            "REQUEST": "GetFeatureInfo",
            "SERVICE": "WMS",
            "VERSION": "1.3.0",
            "INFO_FORMAT": "application/json",
            "FORMAT": "image/png",
            "STYLES": "",
            "TRANSPARENT": "true",
            "QUERY_LAYERS": layer_id,
            "LAYERS": layer_id,
            "WIDTH": tile_size,
            "HEIGHT": tile_size,
            "I": int(i),
            "J": tile_size - int(j),  # отсчет координат для пикселей ведется сверху
            "CRS": "EPSG:4326",  # CRS для bbox
            "BBOX": bbox,
            "FEATURE_COUNT": "10",  # Иначе вернет только один, даже если попало на границу
        }
        response = self.request(
            "get",
            f"/api/aeggis/v3/{layer_id}/wms",
            params=params,  # type: ignore[arg-type]
        )
        return self._validate_feature_collection_response(response)

    def search_at_point(self, pt: Point, layer_def: Type[Feat]) -> Optional[list[Feat]]:
        """Поиск объектов слоя в точке (с типизацией)

        Args:
            pt: Точка поиска
            layer_def: Тип слоя

        Returns:
            Типизированный список объектов, если найдены
        """
        raw_features = self._search_at_point(pt, layer_def.layer_meta.layer_id)
        return self._cast_features_to_layer_defs(raw_features, layer_def)

    def search_at_coords(
        self, lat: float, lng: float, layer_def: Type[Feat]
    ) -> Optional[list[Feat]]:
        """Поиск объектов слоя в координатах

        Args:
            lat: Широта
            lng: Долгота
            layer_def: Тип слоя

        Returns:
            Типизированный список объектов, если найдены
        """
        return self.search_at_point(Point(lng, lat), layer_def)

    ####################
    ### TAB REQUESTS ###
    ####################

    @retry_on_http_error
    def _tab_request(
        self, feat: NspdFeature, tab_class: str, type_: Literal["values", "group"]
    ) -> Optional[dict]:
        if feat.properties.options.no_coords:
            params = {
                "tabClass": tab_class,
                "objdocId": feat.properties.options.objdoc_id,
                "registersId": feat.properties.options.registers_id,
            }
        else:
            params = {
                "tabClass": tab_class,
                "categoryId": feat.properties.category,
                "geomId": feat.id,
            }
        try:
            r = self.request(
                "get", f"/api/geoportal/v1/tab-{type_}-data", params=params
            )
            return r.json()
        except err.NotFound:
            return None

    def _tab_values_request(
        self, feat: NspdFeature, tab_class: str
    ) -> Optional[list[str]]:
        resp = self._tab_request(feat, tab_class, "values")
        if resp is None:
            return None
        return NspdTabResponse.model_validate(resp).value

    def _tab_groups_request(
        self, feat: NspdFeature, tab_class: str
    ) -> Optional[dict[str, Optional[list[str]]]]:
        resp = self._tab_request(feat, tab_class, "group")
        if resp is None:
            return None
        item = NspdTabGroupResponse.model_validate(resp).object
        data = {}
        for i in item:
            title = re.sub(r"[\s:]+$", "", i.title)
            data[title] = i.value
        if len(data) == 0:
            return None
        return data

    def tab_land_parts(self, feat: NspdFeature) -> Optional[list[str]]:
        """Получение данных с вкладки \"Части ЗУ\" """
        return self._tab_values_request(feat, "landParts")

    def tab_land_links(self, feat: NspdFeature) -> Optional[list[str]]:
        """Получение данных с вкладки \"Связанные ЗУ\" """
        return self._tab_values_request(feat, "landLinks")

    def tab_permission_type(self, feat: NspdFeature) -> Optional[list[str]]:
        """Получение данных с вкладки \"Виды разрешенного использования\" """
        return self._tab_values_request(feat, "permissionType")

    def tab_composition_land(self, feat: NspdFeature) -> Optional[list[str]]:
        """Получение данных с вкладки \"Состав ЕЗП\" """
        return self._tab_values_request(feat, "compositionLand")

    def tab_build_parts(self, feat: NspdFeature) -> Optional[list[str]]:
        """Получение данных с вкладки \"Части ОКС\" """
        return self._tab_values_request(feat, "buildParts")

    def tab_objects_list(
        self, feat: NspdFeature
    ) -> Optional[dict[str, Optional[list[str]]]]:
        """
        Получение данных с вкладки \"Объекты\"
        """
        return self._tab_groups_request(feat, "objectsList")

    def get_tab_data(self, feat: NspdFeature, tab_name: TabTitle):
        """Получение данных с указанной вкладки"""
        match tab_name:
            case "Части ЗУ":
                return self.tab_land_parts(feat)
            case "Связанные ЗУ":
                return self.tab_land_links(feat)
            case "Виды разрешенного использования":
                return self.tab_permission_type(feat)
            case "Состав ЕЗП":
                return self.tab_composition_land(feat)
            case "Части ОКС":
                return self.tab_build_parts(feat)
            case "Объекты":
                return self.tab_objects_list(feat)

    #################
    ### SHORTCUTS ###
    #################

    def find_landplot(self, query: str) -> Optional[Layer36048Feature]:
        """Найти ЗУ по кадастровому номеру"""
        return self._filter_search_by_query(self.search_landplots(query), query)

    def find_building(self, query: str) -> Optional[Layer36049Feature]:
        """Найти ОКС по кадастровому номеру"""
        return self._filter_search_by_query(self.search_buildings(query), query)

    def search_landplots_at_point(self, pt: Point) -> Optional[list[Layer36048Feature]]:
        """Поиск ЗУ в точке"""
        return self.search_at_point(pt, Layer36048Feature)

    def search_buildings_at_point(self, pt: Point) -> Optional[list[Layer36049Feature]]:
        """Поиск ОКС в точке"""
        return self.search_at_point(pt, Layer36049Feature)

    def search_landplots(self, cn: str) -> Optional[list[Layer36048Feature]]:
        """Поиск ЗУ по кадастровому номеру"""
        return self.search_in_layer(cn, Layer36048Feature)

    def search_buildings(self, cn: str) -> Optional[list[Layer36049Feature]]:
        """Поиск ОКС по кадастровому номеру"""
        return self.search_in_layer(cn, Layer36049Feature)

    def search_landplots_at_coords(
        self, lat: float, lng: float
    ) -> Optional[list[Layer36048Feature]]:
        """Поиск ЗУ в координатах"""
        return self.search_at_coords(lat, lng, Layer36048Feature)

    def search_buildings_at_coords(
        self, lat: float, lng: float
    ) -> Optional[list[Layer36049Feature]]:
        """Поиск ОКС в координатах"""
        return self.search_at_coords(lat, lng, Layer36049Feature)

    def search_landplots_in_contour(
        self, countour: Union[Polygon, MultiPolygon]
    ) -> Optional[list[Layer36048Feature]]:
        """Поиск ЗУ в контуре"""
        return self.search_in_contour(countour, Layer36048Feature)

    def search_buildings_in_contour(
        self, countour: Union[Polygon, MultiPolygon]
    ) -> Optional[list[Layer36049Feature]]:
        """Поиск ОКС в контуре"""
        return self.search_in_contour(countour, Layer36049Feature)



from typing import (
    Any,
    Generator,
    Literal,
    NoReturn,
    Optional,
    Type,
    TypeVar,
    Union,
    overload,
)

from pynspd.errors import UnknownLayer
from pynspd.map_types._autogen_layers import LayerTitle
from pynspd.schemas import _autogen_features as auto
from pynspd.schemas.base_feature import BaseFeature
from pynspd.schemas.geometries import Geometry
from pynspd.schemas.properties import NspdProperties, OptionProperties

Feat = TypeVar("Feat", bound=Union["BaseFeature", "NspdFeature"])


class NspdFeature(BaseFeature[Geometry, NspdProperties[OptionProperties]]):
    """Базовый класс для валидации GeoJSON-объекта из НСПД"""

    @classmethod
    def _iter_layer_defs(cls) -> Generator[Type[BaseFeature], None, None]:
        root_class = cls.__base__.__base__
        for generic_subclass in root_class.__subclasses__():
            for subclass in generic_subclass.__subclasses__():
                meta = getattr(subclass, "layer_meta", None)
                if meta is not None:
                    yield subclass

    @classmethod
    def by_category_id(cls, category_id: int) -> Type[BaseFeature]:
        """Получение модели по категории"""
        for layer_def in cls._iter_layer_defs():
            if layer_def.layer_meta.category_id == category_id:
                return layer_def
        raise UnknownLayer(category_id)

    @overload
    def cast(
        self, layer_def: None = None
    ) -> BaseFeature[Geometry, NspdProperties[OptionProperties]]: ...

    @overload
    def cast(self, layer_def: Type[Feat]) -> Feat: ...

    def cast(self, layer_def: Optional[Type[Feat]] = None):
        """Приведение объекта к одному из типов перечня определений слоев

        Args:
            layer_def:
                Класс определение слоя. Если не указан, то попытается определить по свойствам. По умолчанию None.

        Raises:
            UnknownLayer: Не удалось определить тип слоя по свойствам

        Returns:
            Объект, приведенный к типу его слоя
        """
        if layer_def is None:
            assert self.properties is not None
            try:
                layer_def = self.by_title(self.properties.category_name)
            except UnknownLayer:
                # скрытый слой, пробуем определить свойства по категории
                similiar_def = self.by_category_id(self.properties.category)
                props_def = similiar_def.model_fields["properties"].annotation
                layer_def = BaseFeature[Geometry, props_def]
        return layer_def.model_validate(self.model_dump(by_alias=True))

    # START_AUTOGEN: title_overload

    @overload
    @classmethod
    def by_title(
        cls, title: Literal["Водная эрозия"]
    ) -> Type[auto.Layer872153Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls, title: Literal["Линейная эрозия"]
    ) -> Type[auto.Layer872155Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls, title: Literal["Ветровая эрозия"]
    ) -> Type[auto.Layer872164Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls, title: Literal["Опустынивание"]
    ) -> Type[auto.Layer872182Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls, title: Literal["Переувлажнение"]
    ) -> Type[auto.Layer872183Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls, title: Literal["Подтопление"]
    ) -> Type[auto.Layer872202Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls, title: Literal["Заболачивание"]
    ) -> Type[auto.Layer872203Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls, title: Literal["Затопление"]
    ) -> Type[auto.Layer872205Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls, title: Literal["Захламление"]
    ) -> Type[auto.Layer872206Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls, title: Literal["Обвально-осыпные и оползневые процессы"]
    ) -> Type[auto.Layer872210Feature]: ...

    @overload
    @classmethod
    def by_title(cls, title: Literal["Абразия"]) -> Type[auto.Layer872211Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls, title: Literal["Кадастровые округа"]
    ) -> Type[auto.Layer36945Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls, title: Literal["Территории объектов культурного наследия"]
    ) -> Type[auto.Layer36316Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls, title: Literal["Кадастровые районы "]
    ) -> Type[auto.Layer36070Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls, title: Literal["Иные нарушенные земли "]
    ) -> Type[auto.Layer872212Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls, title: Literal["Кадастровые кварталы"]
    ) -> Type[auto.Layer36071Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls, title: Literal["Нарушенные земли при наземном строительстве "]
    ) -> Type[auto.Layer872213Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls, title: Literal["Нарушенные земли при гидротехническом строительстве"]
    ) -> Type[auto.Layer872216Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls, title: Literal["Нарушенные земли при недропользовании"]
    ) -> Type[auto.Layer872217Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls, title: Literal["Нарушенные земли при промышленном лесопользовании"]
    ) -> Type[auto.Layer872218Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls, title: Literal["Нарушенные земли при сельскохозяйственном освоении"]
    ) -> Type[auto.Layer872219Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls,
        title: Literal[
            "Нарушенные земли при проведении геологоразведочных, испытательных, эксплуатационных и иных работ"
        ],
    ) -> Type[auto.Layer872220Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls,
        title: Literal[
            "Нарушенные земли при складировании и захоронении промышленных отходов, загрязнение земель"
        ],
    ) -> Type[auto.Layer872221Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls, title: Literal["Государственная граница Российской Федерации"]
    ) -> Type[auto.Layer37313Feature]: ...

    @overload
    @classmethod
    def by_title(cls, title: Literal["Гари"]) -> Type[auto.Layer872222Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls, title: Literal["Негативный процесс отсутствует"]
    ) -> Type[auto.Layer872224Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls, title: Literal["Субъекты Российской Федерации (линии)"]
    ) -> Type[auto.Layer37314Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls, title: Literal["Субъекты Российской Федерации (полигоны)"]
    ) -> Type[auto.Layer37315Feature]: ...

    @overload
    @classmethod
    def by_title(cls, title: Literal["Засоление"]) -> Type[auto.Layer872262Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls, title: Literal["Муниципальные образования (полигональный)"]
    ) -> Type[auto.Layer36278Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls, title: Literal["Муниципальные образования (линейный)"]
    ) -> Type[auto.Layer36279Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls, title: Literal["Населённые пункты (полигоны)"]
    ) -> Type[auto.Layer36281Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls, title: Literal["Населённые пункты (линии)"]
    ) -> Type[auto.Layer37316Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls, title: Literal["Земельные участки из ЕГРН"]
    ) -> Type[auto.Layer36048Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls,
        title: Literal[
            "Земельные участки, образуемые по схеме расположения земельного участка"
        ],
    ) -> Type[auto.Layer37294Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls, title: Literal["Земельные участки, выставленные на аукцион "]
    ) -> Type[auto.Layer37299Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls, title: Literal["Земельные участки, свободные от прав третьих лиц"]
    ) -> Type[auto.Layer37298Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls, title: Literal["Земля для стройки ПКК"]
    ) -> Type[auto.Layer849407Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls, title: Literal["Земля для туризма ПКК"]
    ) -> Type[auto.Layer849453Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls,
        title: Literal["Земельные участки, образуемые по проекту межевания территории"],
    ) -> Type[auto.Layer36473Feature]: ...

    @overload
    @classmethod
    def by_title(cls, title: Literal["Здания"]) -> Type[auto.Layer36049Feature]: ...

    @overload
    @classmethod
    def by_title(cls, title: Literal["Сооружения"]) -> Type[auto.Layer36328Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls, title: Literal["Объекты незавершенного строительства"]
    ) -> Type[auto.Layer36329Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls, title: Literal["Единые недвижимые комплексы"]
    ) -> Type[auto.Layer37433Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls, title: Literal["Предприятие как имущественный комплекс"]
    ) -> Type[auto.Layer37434Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls, title: Literal["ЗОУИТ объектов культурного наследия"]
    ) -> Type[auto.Layer37577Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls, title: Literal["ЗОУИТ объектов энергетики, связи, транспорта"]
    ) -> Type[auto.Layer37578Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls, title: Literal["ЗОУИТ природных территорий"]
    ) -> Type[auto.Layer37580Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls, title: Literal["ЗОУИТ охраняемых объектов и безопасности"]
    ) -> Type[auto.Layer37579Feature]: ...

    @overload
    @classmethod
    def by_title(cls, title: Literal["Иные ЗОУИТ"]) -> Type[auto.Layer37581Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls, title: Literal["Территориальные зоны"]
    ) -> Type[auto.Layer36315Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls, title: Literal["Красные линии "]
    ) -> Type[auto.Layer37293Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls, title: Literal["Особо охраняемые природные территории "]
    ) -> Type[auto.Layer36317Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls, title: Literal["Охотничьи угодья "]
    ) -> Type[auto.Layer36311Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls, title: Literal["Лесничества"]
    ) -> Type[auto.Layer36314Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls, title: Literal["Граница лесопарка"]
    ) -> Type[auto.Layer843763Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls, title: Literal["Береговые линии (границы водных объектов) (полигональный)"]
    ) -> Type[auto.Layer36469Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls, title: Literal["Береговые линии (границы водных объектов)(линейный)"]
    ) -> Type[auto.Layer36470Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls, title: Literal["Особые экономические зоны"]
    ) -> Type[auto.Layer36303Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls, title: Literal["Территории опережающего развития"]
    ) -> Type[auto.Layer36312Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls, title: Literal["Игорные зоны"]
    ) -> Type[auto.Layer36471Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls, title: Literal["Территории выполнения комплексных кадастровых работ"]
    ) -> Type[auto.Layer37430Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls,
        title: Literal[
            "Территория проведения мероприятий по ликвидации накопленного вреда окружающей среде, образовавшегося в результате производства химической продукции в г. Усолье-Сибирское Иркутской области"
        ],
    ) -> Type[auto.Layer37295Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls, title: Literal["Объекты туристского интереса"]
    ) -> Type[auto.Layer849601Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls, title: Literal["Кадастровая стоимость объекта"]
    ) -> Type[auto.Layer37236Feature]: ...

    @overload
    @classmethod
    def by_title(
        cls, title: Literal["Удельный показатель кадастровой стоимости"]
    ) -> Type[auto.Layer37758Feature]: ...

    @overload
    @classmethod
    def by_title(cls, title: Any) -> NoReturn: ...

    # END_AUTOGEN
    @classmethod
    def by_title(cls, title: LayerTitle) -> Type[BaseFeature]:
        """Получение модели слоя по имени"""
        for layer_def in cls._iter_layer_defs():
            if layer_def.layer_meta.title == title:
                return layer_def
        raise UnknownLayer(title)



import httpx
import pytest
from pytest_httpx import HTTPXMock

from pynspd import AsyncNspd
from pynspd.errors import BlockedIP, PynspdResponseError, PynspdServerError


@pytest.mark.asyncio(scope="session")
async def test_429_error(httpx_mock: HTTPXMock, async_api: AsyncNspd):
    httpx_mock.add_response(status_code=429)
    httpx_mock.add_response(status_code=200)
    r = await async_api.safe_request("get", "/api")
    assert r.status_code == 200
    assert len(httpx_mock.get_requests()) == 2


@pytest.mark.asyncio(scope="session")
async def test_400_error(httpx_mock: HTTPXMock, async_api: AsyncNspd):
    httpx_mock.add_response(status_code=400)
    with pytest.raises(PynspdResponseError) as e:
        await async_api.safe_request("get", "/api")
    assert e.value.response.status_code == 400


@pytest.mark.asyncio(scope="session")
async def test_500_error(httpx_mock: HTTPXMock, async_api: AsyncNspd):
    async_api._retries = 1
    httpx_mock.add_response(status_code=500)
    httpx_mock.add_response(status_code=500)
    with pytest.raises(PynspdServerError) as e:
        await async_api.safe_request("get", "/api")
    assert e.value.response.status_code == 500
    assert len(httpx_mock.get_requests()) > async_api._retries


@pytest.mark.asyncio(scope="session")
async def test_remote_disconnect_error(httpx_mock: HTTPXMock, async_api: AsyncNspd):
    httpx_mock.add_exception(httpx.RemoteProtocolError("Unexpected disconnect"))
    httpx_mock.add_response(status_code=200)
    r = await async_api.safe_request("get", "/api")
    assert r.status_code == 200


@pytest.mark.asyncio(scope="session")
async def test_handled_403_error(httpx_mock: HTTPXMock, async_api: AsyncNspd):
    async_api._retry_on_blocked_ip = True
    httpx_mock.add_response(status_code=403)
    httpx_mock.add_response()
    await async_api.safe_request("get", "/api")
    assert len(httpx_mock.get_requests()) == 2


@pytest.mark.asyncio(scope="session")
async def test_unhandled_403_error(httpx_mock: HTTPXMock, async_api: AsyncNspd):
    async_api._retry_on_blocked_ip = False
    httpx_mock.add_response(status_code=403)
    with pytest.raises(BlockedIP):
        await async_api.safe_request("get", "/api")



import os
import re
import ssl
import warnings
from typing import Any, Generator, Optional, Type, TypeVar

import httpx
from hishel import Controller
from hishel._utils import generate_key
from httpcore import Request

from pynspd.errors import AmbiguousSearchError, UnknownLayer
from pynspd.schemas import NspdFeature
from pynspd.schemas.feature import Feat

T = TypeVar("T")


# Нет гарантий, что установлены сертификаты Минцифры, поэтому принудительно отключает проверку ssl
SSL_CONTEXT = ssl._create_unverified_context()
SSL_CONTEXT.set_ciphers("ALL:@SECLEVEL=1")


def _cache_key_generator(request: Request, body: Optional[bytes]) -> str:
    body = body or b""
    key = generate_key(request, body)
    return f"pynspd-{key}"


NSPD_CACHE_CONTROLLER = Controller(
    force_cache=True,
    cacheable_status_codes=[200, 204, 404, 301, 308],
    key_generator=_cache_key_generator,
)


class BaseNspdClient:
    """Базовый класс, не зависящий от sync/async контекста"""

    DNS_HOST = "nspd.gov.ru"
    DNS_URL = "https://" + DNS_HOST

    IP_HOST = "2.63.246.75"
    # random pool?
    # IP_HOST = "2.63.246.76"

    @staticmethod
    def iter_cn(input_str: str) -> Generator[str, None, None]:
        """Извлечение кадастровых номеров из строки"""
        for cn in re.findall(r"\d+:\d+:\d+:\d+", input_str):
            yield cn

    @staticmethod
    def _cast_features_to_layer_defs(
        raw_features: Optional[list[NspdFeature]], layer_def: Type[Feat]
    ) -> Optional[list[Feat]]:
        """Приводит массив фичей к определенному типу"""
        if raw_features is None:
            return None
        features = [i.cast(layer_def) for i in raw_features]
        return features

    @staticmethod
    def _validate_feature_collection_response(
        response: httpx.Response,
    ) -> Optional[list[NspdFeature]]:
        features = response.json()["features"]
        if len(features) == 0:
            return None
        return [NspdFeature.model_validate(i) for i in features]

    @classmethod
    def _filter_search_by_query(
        cls, features: Optional[list[Feat]], query: str
    ) -> Optional[Feat]:
        if features is None:
            return None
        features = list(
            filter(lambda f: query in f.properties.model_dump_json(), features)
        )
        if len(features) > 1:
            features = cls._filter_features_by_query(features, query)
        if len(features) == 0:
            return None
        if len(features) != 1:
            raise AmbiguousSearchError(query)
        return features[0]

    @staticmethod
    def _filter_features_by_query(features: list[Feat], query: str) -> list[Feat]:
        """Поиск точного совпадения поискового запроса в наборе фичей"""

        def in_upper_props(query: str, props: dict[str, Any]) -> bool:
            # если есть в верхнеуровневых свойствах - это точное совпадение
            for v in props.values():
                if v == query:
                    return True
            return False

        def in_option_props(query: str, opts: dict[str, Any]) -> bool:
            # в опциональных свойствах проверяем ключ свойства
            for k, v in opts.items():
                if v == query and "parent" not in k:
                    return True
            return False

        def is_known_category(feat: Feat) -> bool:
            # убеждаемся, что это отображаемая категория
            try:
                feat.cast()  # type: ignore[union-attr]
                return True
            except UnknownLayer:
                return False

        filtered_features = []
        for f in features:
            # иногда поиск дает результат не только по к/н, но и прочим полям
            # например, родительский к/н для помещений
            props = f.properties.model_dump()
            opts = props.pop("options")
            if (
                in_upper_props(query, props) or in_option_props(query, opts)
            ) and is_known_category(f):
                filtered_features.append(f)
        return filtered_features

    @staticmethod
    def _str_var(var_name: str, var: T, trust_env: bool):
        if not trust_env:
            return var
        env_var_name = f"PYNSPD_{var_name.upper()}"
        env_var = os.getenv(env_var_name)
        if env_var is not None:
            if var is not None:
                warnings.warn(
                    f"Используется переменная окружения {env_var_name}, "
                    f"аргумент '{var_name}' проигнорирован",
                    stacklevel=4,
                )
            return env_var
        return var

    @classmethod
    def _int_var(cls, var_name: str, var: T, trust_env: bool):
        d = cls._str_var(var_name, var, trust_env)
        if d is None:
            return None
        return int(d)

    @classmethod
    def _bool_var(cls, var_name: str, var: T, trust_env: bool):
        d = cls._str_var(var_name, var, trust_env)
        if isinstance(d, bool):
            return d
        if isinstance(d, str) and d.lower() in ("true", "1"):
            return True
        return False



from pathlib import Path

import pytest
from typer.testing import CliRunner

from pynspd import __version__
from pynspd.cli import app

runner = CliRunner()


@pytest.fixture()
def output_file(request):
    file: Path = Path.cwd() / ("test" + request.param)
    yield file
    assert file.exists()
    file.unlink()


def test_cn_list_text():
    result = runner.invoke(app, ["search", "77:01:0001057:76"])
    assert result.exit_code == 0


@pytest.mark.parametrize("output_file", [".xlsx"], indirect=True)
def test_cn_list_file(output_file: Path):
    result = runner.invoke(
        app,
        [
            "search",
            "-o",
            output_file.name,
            "-l",
            "./tests/data/cn_list.txt",
        ],
    )
    assert result.exit_code == 0


@pytest.mark.parametrize("output_file", [".csv"], indirect=True)
def test_geo_file(output_file: Path):
    result = runner.invoke(
        app, ["geo", "-o", output_file.name, "./tests/data/poly.gpkg"]
    )
    assert result.exit_code == 0


@pytest.mark.parametrize("output_file", [".gpkg"], indirect=True)
def test_geo_text(output_file: Path):
    result = runner.invoke(
        app, ["geo", "-o", output_file.name, "Point (37.562 55.605)"]
    )
    assert result.exit_code == 0


@pytest.mark.parametrize("output_file", [".geojson"], indirect=True)
def test_geo_coords(output_file: Path):
    result = runner.invoke(
        app,
        [
            "geo",
            "-o",
            output_file.name,
            "   53.193168, 50.106273 53.193704, 50.105026   ",
        ],
    )
    assert result.exit_code == 0


@pytest.mark.parametrize("output_file", [".geojson"], indirect=True)
def test_geo_coords_file(output_file: Path):
    result = runner.invoke(
        app, ["geo", "-o", output_file.name, "./tests/data/pt_list.txt"]
    )
    assert result.exit_code == 0


def test_version():
    result = runner.invoke(app, ["-v"])
    assert result.exit_code == 0
    assert __version__ in result.stdout


def test_bad_text_input():
    result = runner.invoke(app, ["search", "./tests/data/missing_file.txt"])
    assert result.exit_code == 2
    assert "Нужно ввести валидные кадастровые номера" in result.stdout

    result = runner.invoke(app, ["search", "./tests/data/empty_list.txt"])
    assert result.exit_code == 2
    assert "В файле не найдены кадастровые номера" in result.stdout

    result = runner.invoke(app, ["search", "./tests/data/lines.gpkg"])
    assert result.exit_code == 2
    assert "Невозможно прочитать файл" in result.stdout


def test_bad_geom_input():
    result = runner.invoke(app, ["geo", "./tests/data/cn_list.txt"])
    assert result.exit_code == 2
    assert "не является поддерживаемым файлом" in result.stdout

    result = runner.invoke(app, ["geo", "Bad WKT"])
    assert result.exit_code == 2
    assert "не является валидным WKT" in result.stdout

    result = runner.invoke(app, ["geo", "./tests/data/lines.gpkg"])
    assert result.exit_code == 2
    assert "Не поддерживаемый тип геометрии" in result.stdout


def test_empty_output():
    result = runner.invoke(app, ["search", "-p", "fuzzbuzz"])
    assert result.exit_code == 1
    assert "Ничего не найдено" in result.stdout

    result = runner.invoke(
        app, ["geo", "Point (37.55342908811032743 55.59951468019968246)"]
    )
    assert result.exit_code == 1
    assert "Ничего не найдено" in result.stdout


def test_bad_layer_name():
    result = runner.invoke(
        app,
        ["geo", "./tests/data/poly.gpkg", "-c", "---test-layer-name", "Hello World"],
        catch_exceptions=False,
    )
    assert result.exit_code == 2
    assert "не является слоем НСПД" in result.stdout


def test_tab_output():
    result = runner.invoke(app, ["search", "77:07:0006001:1020", "--tab-objects"])
    assert result.exit_code == 0
    assert "Помещения (количество)" in result.stdout
    result = runner.invoke(
        app, ["geo", "Point (37.5106798 55.729467)", "--tab-objects"]
    )
    assert result.exit_code == 0
    assert "Объект недвижимости" in result.stdout


def test_cities_parse():
    result = runner.invoke(
        app,
        [
            "search",
            "-pc",
            "---test-layer-names",
            "Муниципальные образования (полигональный)",
            "---test-layer-names",
            "Населённые пункты (полигоны)",
            "Обнинск",
        ],
    )
    assert result.exit_code == 0
    # два разных субъекта в выводе
    assert "29415000000" in result.stdout
    assert "29415000000" in result.stdout



import re
from collections import Counter
from pathlib import Path
from typing import (
    Annotated,
    Any,
    Callable,
    Generator,
    Optional,
    Sequence,
    Type,
    TypeVar,
    get_args,
)

import geopandas as gpd
import questionary
import typer
from pyogrio.errors import DataSourceError
from rich import print
from rich.progress import (
    BarColumn,
    Progress,
    SpinnerColumn,
    TaskProgressColumn,
    TextColumn,
    TimeRemainingColumn,
)
from shapely import GEOSException, MultiPolygon, Point, Polygon, from_wkt

from pynspd import Nspd, NspdFeature, __version__
from pynspd.errors import TooBigContour, UnknownLayer
from pynspd.map_types._autogen_layers import LayerTitle
from pynspd.schemas import BaseFeature

app = typer.Typer(
    pretty_exceptions_show_locals=False,
    no_args_is_help=True,
    help="Утилита командной строки для поиска на НСПД",
)
CN_PATTERN = re.compile(r"\d+:\d+:\d+:\d+")


T = TypeVar("T")


OutputOption = Annotated[
    Optional[Path],
    typer.Option(
        "--output",
        "-o",
        help=(
            "Файл, в который будет сохранен результат поиска. "
            "Поддерживаются гео- (.gpkg, .geojson и пр.) и табличные форматы (.xlsx, .csv)"
        ),
        rich_help_panel="General Options",
    ),
]

LocalizeOption = Annotated[
    bool,
    typer.Option(
        "--localize",
        "-l",
        help="Использовать названия колонок с сайта, а не из оригинальные",
        rich_help_panel="General Options",
    ),
]

TabObjectsOption = Annotated[
    bool,
    typer.Option(
        "--tab-objects",
        help='Получить данные со вкладки "Объекты" для найденных объектов',
        rich_help_panel="General Options",
    ),
]

PlainOption = Annotated[
    bool,
    typer.Option(
        "--plain",
        "-p",
        help="Не конвертировать входные данные в список к/н",
        rich_help_panel="General Options",
    ),
]

ChooseLayersOption = Annotated[
    bool,
    typer.Option(
        "--choose-layer",
        "-c",
        help="Выбрать слои из списка вместо слоя по умолчанию",
        rich_help_panel="General Options",
    ),
]


def define_queries(input_: str, plain_query: bool) -> list[str]:
    """Поиск массива к/н в исходной строке или текстовом файле"""
    maybe_file = Path(input_)
    if maybe_file.exists():
        try:
            with open(maybe_file, encoding="utf-8", mode="r") as file:
                content = file.read()
        except UnicodeDecodeError:
            raise typer.BadParameter(f"Невозможно прочитать файл {input_}")
        if plain_query:
            return content.splitlines()
        cns = list(Nspd.iter_cn(content))
        if len(cns) == 0:
            raise typer.BadParameter(
                "В файле не найдены кадастровые номера. Используйте флаг --plain / -p для поиска по обычному тексту"
            )
        return cns
    if plain_query:
        return [input_]
    if not CN_PATTERN.match(input_):
        raise typer.BadParameter(
            "Нужно ввести валидные кадастровые номера или использовать флаг --plain / -p для обычного текста"
        )
    return list(Nspd.iter_cn(input_))


def define_geoms(input_: str) -> list[Point] | list[Polygon] | list[MultiPolygon]:
    """Проверка гео-файла и получение геометрии из него"""
    maybe_file = Path(input_)
    if maybe_file.exists() and maybe_file.suffix == ".txt":
        pts_text = maybe_file.read_text("utf-8")
    else:
        pts_text = input_
    pts_text = pts_text.strip()
    matches = re.findall(r"(\d{2,3}\.\d+),? *(\d{2,3}\.\d+)", pts_text)
    if re.match(r"^\d", pts_text) and len(matches) > 0:
        coords = [list(map(float, i)) for i in matches]
        pts = [Point(*i[::-1]) for i in coords]
        geoseries = gpd.GeoSeries(pts, crs=4326)
    elif maybe_file.exists():
        try:
            geoseries = gpd.read_file(input_).geometry
        except DataSourceError as e:
            raise typer.BadParameter(
                f"{input_} не является поддерживаемым файлом"
            ) from e
    else:
        try:
            wkt_geom = from_wkt(input_)
            geoseries = gpd.GeoSeries([wkt_geom], crs=4326)
        except GEOSException as e:
            raise typer.BadParameter(
                f"Файл {input_} не существует или не является валидным WKT"
            ) from e
    geom_types = geoseries.geom_type.unique().tolist()
    geometry = geoseries.tolist()
    if len(geom_types) > 1:
        raise typer.BadParameter("Не поддерживаются файла со смешанной геометрией")
    if geom_types[0] not in ("Point", "Polygon", "MultiPolygon"):
        raise typer.BadParameter(f"Не поддерживаемый тип геометрии - {geom_types[0]}")
    return geometry


def define_layer_def(layer_name: str) -> Type[BaseFeature]:
    """Определение типа слоя"""
    try:
        return NspdFeature.by_title(layer_name)
    except UnknownLayer as e:
        raise typer.BadParameter(f"{layer_name} не является слоем НСПД") from e


def _progress_iter(items: Sequence[T]) -> Generator[T, None, None]:
    cols = [
        SpinnerColumn(),
        TextColumn("{task.description}"),
    ]
    if len(items) > 1:
        cols += [
            BarColumn(),
            TaskProgressColumn(),
            TimeRemainingColumn(),
        ]
    with Progress(
        *cols,
        transient=True,
    ) as progress:
        task = progress.add_task(description="Поиск...", total=len(items))
        for item in items:
            yield item
            progress.update(task, advance=1)


def _get_features_from_list(
    client: Nspd,
    queries: list[str],
    layer_defs: Optional[list[Type[BaseFeature]]],
) -> Optional[list[NspdFeature]]:
    features = []
    missing = []
    for query in _progress_iter(queries):
        if layer_defs is None:
            feats = client.search(query)
        else:
            feats = client.search_in_layers(query, *layer_defs)
        if feats is None:
            missing.append(query)
            continue
        features += feats
    if missing:
        print(
            f":warning-emoji: [orange3] Не найдены {len(missing)} из {len(queries)} объектов:"
        )
        for m in missing:
            print(f"[orange3]   - {m}")
        print()
    if len(features) == 0:
        return None
    return features


def _get_features_from_geom(
    method: Callable[
        [Point | Polygon | MultiPolygon, Type[BaseFeature]], Optional[list[BaseFeature]]
    ],
    geoms: list[Point | Polygon | MultiPolygon],
    layer_def: Type[BaseFeature],
) -> Optional[list[BaseFeature]]:
    features = []
    missing_count = 0
    for geom in _progress_iter(geoms):
        feats = method(geom, layer_def)
        if feats is None:
            missing_count += 1
            continue
        features += feats
    if missing_count:
        print(
            f":warning-emoji: [orange3] Ничего не найдено в {missing_count} из {len(geoms)} локаций"
        )
        print()
    if len(features) == 0:
        return None
    return features


def _get_tab_object(client: Nspd, features: list[NspdFeature]) -> list[NspdFeature]:
    """Получение данных из вкладки"""
    for feat in _progress_iter(features):
        data = client.tab_objects_list(feat)
        if data is None:
            continue
        assert feat.properties.options.model_extra is not None
        feat.properties.options.model_extra["tab"] = data
    return features


def prepare_features(features: list[NspdFeature], localize: bool) -> gpd.GeoDataFrame:
    prepared_features: list = []
    unknown_layers = Counter()
    for feat in features:
        assert feat.properties.options.model_extra is not None
        props: dict[str, Any] = feat.properties.options.model_extra.pop("tab", {})
        if localize:
            try:
                feat = feat.cast()
            except UnknownLayer:
                unknown_layers[feat.properties.category_name] += 1
                continue
            props.update(feat.properties.options.model_dump_human_readable())
            props["Категория"] = feat.properties.category_name
        else:
            props.update(feat.properties.options.model_dump())
            props["category"] = feat.properties.category_name
        props["geometry"] = feat.geometry.to_shape()
        prepared_features.append(props)
    for category, count in unknown_layers.items():
        print(
            f'[orange3] Было пропущено объектов с неизвестным слоем "{category}" - {count} (невозможно локализировать данные)'
        )
    return gpd.GeoDataFrame(prepared_features, crs=4326).fillna("")


def process_output(
    features: Optional[list[NspdFeature]], output: Optional[Path], localize: bool
) -> None:
    if features is None:
        print("[red]Ничего не найдено")
        raise typer.Abort()
    gdf = prepare_features(features, localize)
    try:
        if output is None:
            print(gdf.T)
            return
        elif output.suffix == ".xlsx":
            gdf.to_excel(output, index=False)
        elif output.suffix == ".csv":
            gdf.to_csv(output, index=False)
        else:
            gdf.to_file(output)
    except PermissionError:
        _ = typer.prompt(
            f"Файл {output} открыт в другой программе. Закройте его и нажмите любую клавишу",
            default=True,
            hide_input=True,
            prompt_suffix="...",
            show_default=False,
        )
        return process_output(features, output, localize)
    print(f"[green]Найдено {len(gdf)} объектов, сохранено в файл {output.resolve()}[/]")


def version_callback(version: bool = False) -> None:
    if version:
        print(__version__)
        raise typer.Exit()


@app.callback()
def common(
    version: Annotated[
        bool,
        typer.Option(
            "-v",
            "--version",
            help="Show current version",
            is_eager=True,
            callback=version_callback,
        ),
    ] = False,
):
    pass


@app.command(no_args_is_help=True)
def geo(
    input: Annotated[
        str,
        typer.Argument(
            help="Путь к файлу с геоданными, координаты точек или WKT",
            show_default=False,
        ),
    ],
    choose_layer: ChooseLayersOption = False,
    output: OutputOption = None,
    localize: LocalizeOption = False,
    add_tab_object: TabObjectsOption = False,
    _test_layer_name: Annotated[Optional[str], typer.Option(hidden=True)] = None,
) -> None:
    """Поиск объектов по геоданным

    Может производить поиск по файлам gpkg, geojson, координатам точек или WKT-строке.
    По умолчанию ищет в слое "Земельные участки из ЕГРН".
    """
    if choose_layer:
        layer_name = (
            _test_layer_name
            if _test_layer_name is not None
            else questionary.select(
                "Выберите слой: ", choices=get_args(LayerTitle)
            ).ask()
        )
        if layer_name is None:
            raise typer.Abort()
    else:
        layer_name = "Земельные участки из ЕГРН"
    geoms = define_geoms(input)
    layer_def = define_layer_def(layer_name)
    with Nspd() as client:
        if isinstance(geoms[0], Point):
            features = _get_features_from_geom(client.search_at_point, geoms, layer_def)
        else:
            try:
                features = _get_features_from_geom(
                    client.search_in_contour, geoms, layer_def
                )
            except TooBigContour:
                raise typer.BadParameter(
                    "НСПД не может ответить из-за слишком большой площади поиска"
                )
        if features and add_tab_object:
            _get_tab_object(client, features)
    process_output(features, output, localize)


@app.command(no_args_is_help=True)
def search(
    input: Annotated[
        str,
        typer.Argument(
            help="Поисковой запрос. Может быть многострочным текстовым файлом",
            show_default=False,
        ),
    ],
    choose_layer: ChooseLayersOption = False,
    plain_query: PlainOption = False,
    output: OutputOption = None,
    localize: LocalizeOption = False,
    add_tab_object: TabObjectsOption = False,
    _test_layer_names: Annotated[Optional[list[str]], typer.Option(hidden=True)] = None,
) -> None:
    """Поиск объектов по тексту

    По умолчанию разбивает запрос на кадастровые номера и ищет в объектах недвижимости.
    """
    if choose_layer:
        layer_names = (
            _test_layer_names
            if _test_layer_names is not None
            else questionary.checkbox(
                "Выберите слои: ",
                choices=get_args(LayerTitle),
                validate=lambda x: "Не выбрано ни одно значение"
                if len(x) == 0
                else True,
            ).ask()
        )
        if layer_names is None:
            raise typer.Abort()
        layer_defs = [define_layer_def(layer_name) for layer_name in layer_names]
    else:
        layer_defs = None
    queries = define_queries(input, plain_query)
    with Nspd() as client:
        features = _get_features_from_list(client, queries, layer_defs)
        if features and add_tab_object:
            features = _get_tab_object(client, features)
    process_output(features, output, localize)


def main() -> Any:
    return app()
